{}
{                                                       }
{      Virtual Pascal for Linux                         }
{      PE to ELF conversion utility                     }
{      }
{      Copyright (C) 1999 Veit Kannegieser              }
{                    2008 Angels Holocaust              }
{                                                       }
{}

(*$Use32+,I+,Delphi+,X+*)
(*  $D+,L+,Locinfo+*)
(*$M 1000000*)

program pe2elf;

// 1999.07.28 Veit Kannegieser
// ... VK/JP
// 1999.10.22 Version mit funkionierendem .so-Import an J.P. verschickt
// 2000.05.10 -v fr Sektionenliste und word->smallword fr PE2LETY.PAS
// 2000.12.17 realloc const32/4096, ld-linux.so.2, exe-beschreibung
// 2001.02.25 interpreter nur verwenden wenn importe vorhanden sind
//            -> sleep aus libc_6.so verwenden
// 2001.05.28 Resourcefehler behoben
// 2001.08.12 nicht benutzte Resourcen verwerfen
// 2001.12.09 FreeBSD brandelf-Funktionalitt eingebaut
// 2001.12.09 Unicode-Zeichensatz-Konverter mit und ohne Unit ZST (OnlyCP819)
// 2003.08.06 use pBufStream, use Unicode/Resource converter from pe2le
//            use some FillChar after ReallocMem to avaoid garbage in executable (ReallocMem does not zerofill!)
(*
   13.07.2008 - uses clause now lowercase for compilation in LINUX
              - added PAGESIZE constant
              - FreeBSD compatibility:
                * program headers come before the section headers
                  from imgact_elf.c:
                  /* Only support headers that fit within first page for now      */
                  /*    (multiplication of two Elf_Half fields will not overflow) */
                  if ((hdr->e_phoff > PAGE_SIZE) ||
                      (hdr->e_phentsize * hdr->e_phnum) > PAGE_SIZE - hdr->e_phoff) {
                      error = ENOEXEC;
                      goto fail;
                  }
   16.07.2008 - Added version numbering and own date output
              - Included impconve.pas proghead.pas search_d.pas directly into the file
              - Renamed: Aufrunden() -> RoundUp(), sektionen_namen -> SectionNames,
                         dateiposition -> FilePosition, AdressPosition -> AddressPosition
              - added PAGESIZE constant
              - put the parameter parsing into ParseCommandLineOption() for recursion
              - Added new commandline options: -F -A
              - FreeBSD compatibility:
                * program headers come before the section headers
                * sections & headers are aligned to 4 bytes
   17.07.2008 - Added a new table, which holds all information for one ELF part
              - Renamed: sektionen -> tSections
              - Extended tSections to hold all possible ELF parts (header, description, ...)
              - Replaced SectionNames[] + sektionen_uebersicht[] with SectionTable[]
              - Moved speicher_tabelle/programm_kopf/sektionen_kopf into SectionTable[]
              - Added SectionSortArray[] and sorting after p_vaddr
   18.07.2008 - added FinishSections() and SectionEntry.Has_sh/ph
   19.07.2008 - Rewrote the section generation code in rechne():
                * now using a loop for sh/ph filling, which allows free ordering of
                  sections
                * Removed bearbeite_programmheader()
              - Rewrote the section saving code in schreibe_elf():
                * replaced schreibe_elf().laenge with SectionTable[].Size
              - Added sekt_(sect|prog)_headers/sekt_elf_header usage
   29.07.2008 - added sekt_desc
   27.11.2008 - Improved Has_(ph|sh) handling: not pre-initialized anymore
              - Removed the e_phnum/programm_kopf_anzahl error exception

   12.04.2009 - New PE-file loader from PE2LE with many linkers support.
                Search VPT in remarks.

   TODO: - convert/restructure existing (old PE2ELF) ELF files into new
           PE2ELF format
         - .note section: elf doku seite 42: note
         - make tSections and SectionTable[] ordering independent
*)

uses
  pe2lety,
  reskonv,
  descript,
  elfdef,
  dbwrite,
  strings,
  dos,
  objects,
  vputils,
  vpsyslow;

const
  datum                         ='2003.08.06';
  ah_date                       ='28.11.2008';
  updated                       ='* updated by VPT * 12.04.2009';
  version                       ='v1.20';
  memorymode                    :word={1}3;
  include_section_header        :boolean=true;
  resource_codepage             :string='ISO-8859-1';
  linebreak                     :array[1..2] of char=(^m,^j);
  descriptionframechar          :char='-';

  elf_kopf:Elf32_Ehdr_type=
   (magic0123                   :magic0123_elf;
    file_class                  :file_class__32bit;
    data_encoding               :ELFDATA2LSB;
    file_version                :e_version__current;
    osabi                       :ELFOSABI_NONE;
    abi_version                 :0;
    padding                     :(0,0,0,0,0,0,0);
    e_type                      :e_type__exec;
    e_machine                   :e_machine__intel386;
    e_version                   :e_version__current;
    e_entry                     :0;                     // entrypoint
    e_phoff                     :0;                     // program header offset
    e_shoff                     :0;                     // sections header offset
    e_flags                     :0;                     // keine
    e_ehsize                    :SizeOf(Elf32_Ehdr_type);
    e_phentsize                 :SizeOf(program_header_type);
    e_phnum                     :0;                     // 0..e_phnum-1 of entrys
    e_shentsize                 :SizeOf(section_header_type);
    e_shnum                     :0;                     // 0..e_shnum-1 of entrys
    e_shstrndx                  :0);

  delta=0;

  PAGESIZE    = 4096;
  BASEADDRESS = $00401000; // $00401000 damit der VP-debugger funktioniert
  Alignment : Longint = 4;

type
  PCharArray                    = ^TCharArray;
  TCharArray                    = array[0..512*1024*1024      ] of Char;

  dynamic_tabelle_typ           =array[0..20000] of Elf32_Dyn_type;
  dynamic_tabelle_z_typ         =^dynamic_tabelle_typ;

  symboltabelle_typ             =array[0..200000] of Elf32_Sym_type;

  rel_plt_typ                   =array[0..200000] of Elf32_Rel_type;
  rel_plt_z_typ                 =^rel_plt_typ;

var
  d1,d2                         :pBufStream;
  pe_kopf                       :pe_kopf_typ;
  pe_obj_tab                    :array[pe_obj_typen] of pe_obj_typ;

  code_puffer,
  bss_puffer,
  data_puffer,
  import_puffer,
  relo_puffer,
  reso_temp_puffer,
  reso_puffer                   :pByteArray;

  plt_speicher                  :pByteArray;
  got_speicher                  :pLongArray;
  hash_tabelle                  :pLongArray;

  symboltabelle                 :^symboltabelle_typ;

  symbolstringtabelle           :PCharArray;

  rel_plt_speicher              :rel_plt_z_typ;

  importe_vorhanden,
  resourcen_vorhanden,
  bss_vorhanden                 :boolean;
  code_vorhanden, data_vorhanden,
  relo_vorhanden               :boolean;

  sektionen_kopf                :array[0..40] of section_header_type;
  pk2                           :array[0..40] of program_header_type;

  string_puffer                 :pByteArray;
  string_puffer_laenge          :longint;
  reso_temp_laenge              :longint;
  anzahl_importe                :longint;

  InFile, OutFile               :string;

  programmlader                 :array[0..255] of char;

type
  {AH 17.07.2008 Sections marked with (*) are no official sections and just added
                 so this enum can be used to handle all parts of an ELF file, not just
                 the officially defined sections; this also sets the order of the entries
                 in the ELF file}
  tSections=
    (
     sekt_elf_header,   // the ELF-Fileheader (*)
     sekt_desc,         // Custom module/file description (*)
     sekt_note,         // Note section (TODO)
     sekt_prog_headers, // Program headers (*)
     sekt_sect_headers, // Section headers (*)
     sekt_null,         // reserviert
     sekt_interpreter,  // Programmlader
     sekt_text,         // Code
     sekt_bss,          // wenn vorhanden
     sekt_data_a,       // \ Konstanten, / alle initialisierten Seiten
     sekt_data_b,       // / Variablen   \ alle Nullseiten
     sekt_dynamic,      // Informationen/Zeiger fr den Programmlader (Dynamic Linking Information)
     sekt_got,          // Adressen importierter Funktionen (Global Offset Table)
     sekt_dynsym,       // Tabelle importierte Symbole
     sekt_dynsymstr,    // Namen fr sekt_dynsym
     sekt_dynsymstrhash,// Hash-Tabelle fr sekt_dynsymstr (Symbol Hash Table)
     sekt_resource,     // Resourcen
     sekt_rel_plt,      // Verbindung von plt zu dynsym
     sekt_plt,          // Aufrufe von importierten Funktionen (Procedure Linkage Table)
     sekt_strings       // Namen von Sektionen
    );

  {AH 17.07.2008 Record which holds all information needed to handle a section}
  SectionEntry = packed record
                   Id     : tSections;
                   Name   : string[10];
                   Desc   : string[25];
                   Exists : Boolean;
                   ph     : program_header_type;
                   sh     : section_header_type;
                   Index  : Longint;
                   Address: Pointer;
                   {Filled at runtime}
                   Size   : Longint;
                   Has_ph : Boolean;
                   Has_sh : Boolean;
                 end;

const
//  SectionTable: Array[low(tSections)..high(tSections)] of SectionEntry =
  SectionTable: Array[tSections] of SectionEntry =
    (
     (Id: sekt_elf_header   ; Name:'*elf_head'; Desc:'ELF header'; Exists: True;
                              ph: ( );
                              sh: ( );
                              Index: 0;
                              Address: @elf_kopf
     ),
     (Id: sekt_desc        ; Name:'*desc'; Desc:'ELF description'; Exists: True;
                              ph: ( );
                              sh: ( );
                              Index: 0;
                              Address: @module_description[1]
     ),
     (Id: sekt_note         ; Name:'.note'; Desc:'Note section'; Exists: False;
                              ph: (
                                    p_type      : PT_NOTE;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R;
                                    p_align     : 1
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_NOTE;
                                    sh_flags    : 0;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: 1;
                                    sh_entsize  : 0
                                  );
                              Index: 0;
                              Address: nil
     ),
     (Id: sekt_prog_headers ; Name:'*PRG_HEAD'; Desc:'Program headers'; Exists: True;
                              ph: (
                                    p_type      : PT_PHDR;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R;
                                    p_align     : 4
                                  );
                              sh: ( );
                              Index: 0;
                              Address: @pk2
     ),
     (Id: sekt_sect_headers ; Name:'*SEC_HEAD'; Desc:'Section headers'; Exists: False;
                              ph: ( );
                              sh: ( );
                              Index: 0;
                              Address: @sektionen_kopf
     ),
     (Id: sekt_null   {Null}; Name:''         ; Desc:'null,reserved'; Exists: True;
                              ph: ( );
                              sh: ( )
     ),
     (Id: sekt_interpreter  ; Name:'.interp'  ; Desc:'interpreter'; Exists: False;
                              ph: (
                                    p_type      : PT_INTERP;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R;
                                    p_align     : 1
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_PROGBITS;
                                    sh_flags    : SHF_ALLOC;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: 1;
                                    sh_entsize  : 0
                                  )
     ),
     (Id: sekt_text   {Code}; Name:'.text'    ; Desc:'.text (code)'; Exists: True ;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_X+PF_R;
                                    p_align     : PAGESIZE
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_PROGBITS;
                                    sh_flags    : SHF_ALLOC+SHF_EXECINSTR;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: {4 PAGESIZE}16;
                                    sh_entsize  : 0
                                  )
     ),
     (Id: sekt_bss          ; Name:'.bss'     ; Desc:'"_bss"'; Exists: False;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R+PF_W;
                                    p_align     : 4
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_NOBITS;
                                    sh_flags    : SHF_ALLOC+SHF_WRITE;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: 4;
                                    sh_entsize  : 0
                                  )
     ),
     (Id: sekt_data_a{Const}; Name:'.data'    ; Desc:'data_a (.data)'; Exists: True;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R+PF_W;
                                    p_align     : PAGESIZE
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_PROGBITS;
                                    sh_flags    : SHF_ALLOC+SHF_WRITE;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: 4;
                                    sh_entsize  : 0
                                  )
     ),
     (Id: sekt_data_b       ; Name:'.bss'     ; Desc:'data_b (.bss)'; Exists: False;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R+PF_W;
                                    p_align     : 4
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_NOBITS;
                                    sh_flags    : SHF_ALLOC+SHF_WRITE;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: 4;
                                    sh_entsize  : 0
                                  )
     ),
     (Id: sekt_dynamic      ; Name:'.dynamic' ; Desc:'.dynamic'; Exists: False;
                              ph: (
                                    p_type      : PT_DYNAMIC;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R+PF_W;
                                    p_align     : 4
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_DYNAMIC;
                                    sh_flags    : SHF_WRITE+SHF_ALLOC;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : ord(sekt_dynsymstr);
                                    sh_info     : 0;
                                    sh_addralign: 4;
                                    sh_entsize  : SizeOf(Elf32_Dyn_type)
                                  )
     ),
     (Id: sekt_got          ; Name:'.got'     ; Desc:'.got'; Exists: False;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R+PF_W;
                                    p_align     : 4
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_PROGBITS;
                                    sh_flags    : SHF_WRITE+SHF_ALLOC;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: 4;
                                    sh_entsize  : 4
                                  )
     ),
     (Id: sekt_dynsym       ; Name:'.dynsym'  ; Desc:'.dynsym'; Exists: False;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R;
                                    p_align     : 4
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_DYNSYM;
                                    sh_flags    : SHF_ALLOC;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : ord(sekt_dynsymstr);
                                    sh_info     : ord(sekt_interpreter);
                                    sh_addralign: 4;
                                    sh_entsize  : SizeOf(Elf32_Sym_type)
                                  )
     ),
     (Id: sekt_dynsymstr    ; Name:'.dynstr'  ; Desc:'.dynsymstr'; Exists: False;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R;
                                    p_align     : 1
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_STRTAB;
                                    sh_flags    : SHF_ALLOC;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: 1;
                                    sh_entsize  : 0
                                  )
     ),
     (Id: sekt_dynsymstrhash; Name:'.hash'    ; Desc:'.dynsymstrhash'; Exists: False;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R;
                                    p_align     : 4
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_HASH;
                                    sh_flags    : SHF_ALLOC;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : ord(sekt_dynsym);
                                    sh_info     : 0;
                                    sh_addralign: 4;
                                    sh_entsize  : 4
                                  )
     ),
     (Id: sekt_resource     ; Name:'resource' ; Desc:'resource'; Exists: False;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R;
                                    p_align     : 4
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_PROGBITS;
                                    sh_flags    : SHF_ALLOC;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: 4;
                                    sh_entsize  : 0
                                  )
     ),
     (Id: sekt_rel_plt      ; Name:'.rel.plt' ; Desc:'.rel.plt'; Exists: False;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R;
                                    p_align     : 4
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_REL;
                                    sh_flags    : SHF_ALLOC;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : ord(sekt_dynsym);
                                    sh_info     : ord(sekt_plt);
                                    sh_addralign: 4;
                                    sh_entsize  : SizeOf(Elf32_Rel_type)
                                  )
     ),
     (Id: sekt_plt          ; Name:'.plt'     ; Desc:'.plt'; Exists: False;
                              ph: (
                                    p_type      : PT_LOAD;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R+PF_X;
                                    p_align     : 4
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_PROGBITS;
                                    sh_flags    : SHF_ALLOC+SHF_EXECINSTR;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: 4;
                                    sh_entsize  : 0
                                  )
     ),
     (Id: sekt_strings      ; Name:'.shstrtab'; Desc:'sektion header strings'; Exists: False;
                              ph: (
                                    p_type      : PT_NULL;
                                    p_offset    : 0;
                                    p_vaddr     : 0;
                                    p_paddr     : 0;
                                    p_filesz    : 0;
                                    p_memsz     : 0;
                                    p_flags     : PF_R;
                                    p_align     : 1
                                  );
                              sh: (
                                    sh_name     : 0;
                                    sh_type     : SHT_STRTAB;
                                    sh_flags    : 0;
                                    sh_addr     : 0;
                                    sh_offset   : 0;
                                    sh_size     : 0;
                                    sh_link     : 0;
                                    sh_info     : 0;
                                    sh_addralign: 1;
                                    sh_entsize  : 0
                                  )
     )
    );

  {Alte ELF-Inhalt-Reihenfolge: ELF-header, description, section headers, program headers}
  {Neue ELF-Inhalt-Reihenfolge: ELF-header, description, program headers, section headers}

var
  SectionSortArray              : Array [tSections] of tSections;
  SectionSortArrayElementCount  : tSections;
  anzahl_dynamic_eintraege      :longint;
  dynamic_speicher              :dynamic_tabelle_z_typ;

procedure Abbruch(const zk:string);
  begin
    WriteLn(zk);
    Halt(255);
  end;

procedure RoundUp(var a:longint;const m:longint);
  var
    rest:longint;
  begin
    rest:=a mod m;
    if rest<>0 then
      a:=a+m-rest;
  end;

//

procedure OpenReadBST(var f:pBufStream;const filename:string);
  begin
    f:=New(pBufStream,Init(filename,stOpenRead,512*1024));
    if f^.Status<>stOK then
      Abbruch('Cannot open'+filename+'!');
  end;

procedure CreateBST(var f:pBufStream;const filename:string);
  begin
    f:=New(pBufStream,Init(filename,stCreate,512*1024));
    if f^.Status<>stOK then
      Abbruch('Error creating'+filename+'!');
  end;

procedure BlockReadBST(f:pBufStream;var buffer;const size:longint);
  begin
    f^.Read(buffer,size);
    if f^.Status<>stOK then
      Abbruch('Source stream read error');
  end;

procedure BlockWriteBST(f:pBufStream;var buffer;const size:longint);
  begin
    f^.Write(buffer,size);
    if f^.Status<>stOK then
      Abbruch('Target stream write error');
  end;

procedure SeekBST(f:pBufStream;l:longint);
  begin
    f^.Seek(l);
    if f^.Status<>stOK then
      Abbruch('Stream seek error');
  end;

procedure CloseBST(var f:pBufStream);
  begin
    f^.Done;
    f:=nil;
  end;

function FileSizeBST(f:pBufStream):longint;
  begin
    Result:=f^.GetSize;
    if f^.Status<>stOK then
      Abbruch('Stream FileSize error');
  end;

procedure lade_pe;



  var
    datei,
    res_datei           :pBufStream;
    puffer              :array[0..$3f] of char;
    p1                  :longint;
    i                   :SmallWord;
    new_size, fill_size :longint;
    pe_start            :longint;
    pe_obj_tab_all                                :array[0..95] of pe_obj_typ;

  procedure lies(var p:pByteArray;var obj:pe_obj_typ);
    begin
      GetMem(p,obj.physical_size);
      SeekBST(datei,obj.physical_offset);
      BlockReadBST(datei,p^,obj.physical_size);
    end;

  function gleich4(var v1; const v2: array of char): boolean;
  var
    x1: longint absolute v1;
    x2: longint absolute v2;
  begin
    gleich4:= x1 = x2;
  end;

  begin
    OpenReadBST(datei,InFile);

    pe_start:=0;
    repeat
      BlockReadBST(datei,puffer,SizeOf(puffer));

      if gleich4(puffer, 'PE'#0#0) then break;
      if gleich4(puffer, 'PL'#0#0) then break; // VPT: Pharlap signature
      if gleich4(puffer, 'PX'#0#0) then break; // VPT: HX and DOS4G signature
      if gleich4(puffer, 'BPE'#0)  then break; // VPT: Unilink/Doswin32 signature

      if ((puffer[0]='M') and (puffer[1]='Z'))
      or ((puffer[0]='Z') and (puffer[1]='M'))
       then
        begin
          pe_start:=pLongint(@puffer[$3c])^;
          SeekBST(datei,pe_start);
          Continue;
        end;

      if gleich4(puffer, 'LX'#0#0) then
        Abbruch('Source is LX!');

      if gleich4(puffer, 'LE'#0#0) then
        Abbruch('Source is LE!');

      Abbruch('Unknown format!');

    until false;

    SeekBST(datei,pe_start);
    BlockReadBST(datei,pe_kopf,SizeOf(pe_kopf));

    (* obj-Tabelle laden *)

    if pe_kopf.num_obj > 96 then
      Abbruch('Schade!');

    SeekBST(datei,pe_start+pe_kopf.nthdrsize+$18);
    for i:=0 to pe_kopf.num_obj - 1 do
      BlockReadBST(datei,pe_obj_tab_all[i],SizeOf(pe_obj_typ));

    pe_obj_tab[code32 ].name            := '.text'#0#0#0;
    pe_obj_tab[code32 ].virtual_size    := 0;
    pe_obj_tab[code32 ].rva             := pe_kopf.text_start;
    pe_obj_tab[code32 ].physical_size   := 0;
    pe_obj_tab[code32 ].physical_offset := 0;
    pe_obj_tab[code32 ].obj_flags       := scnCode;
    GetMem(code_puffer, pe_kopf.tsize);

    pe_obj_tab[const32].name            := '.data'#0#0#0;
    pe_obj_tab[const32].virtual_size    := 0;
    pe_obj_tab[const32].rva             := pe_kopf.data_start;
    pe_obj_tab[const32].physical_size   := 0;
    pe_obj_tab[const32].physical_offset := 0;
    pe_obj_tab[const32].obj_flags       := scnData;
    GetMem(data_puffer, pe_kopf.dsize);

    for i:=0 to pe_kopf.num_obj - 1 do
      begin
        {VPT: code section found}
        if ((pe_obj_tab_all[i].obj_flags and scnCode)=scnCode)
           (* and (pe_obj_tab_all[i].rva>=pe_kopf.text_start)
           and (pe_obj_tab_all[i].physical_size>0)
           and (pe_obj_tab_all[i].virtual_size>0) *)
           then
          begin
            if pe_kopf.text_start = 0 then {VPT: alink/blink bug}
              begin
                pe_kopf.text_start      := pe_obj_tab_all[i].rva;
                pe_obj_tab[code32 ].rva := pe_obj_tab_all[i].rva
              end;
            new_size := pe_obj_tab_all[i].physical_size
                     +  (pe_obj_tab_all[i].rva-pe_kopf.text_start);
            ReAllocMem(code_puffer,new_size);
            SeekBST(datei,pe_obj_tab_all[i].physical_offset);
            BlockReadBST
              (
                datei,
                code_puffer^[pe_obj_tab_all[i].rva-pe_kopf.text_start],
                pe_obj_tab_all[i].physical_size
              );
            fill_size := pe_obj_tab_all[i].rva
              - (pe_obj_tab[code32 ].rva + pe_obj_tab[code32 ].physical_size);
            if fill_size > 0 then
              FillChar(code_puffer^[pe_obj_tab[code32 ].physical_size], fill_size, $CC);
            pe_obj_tab[code32 ].physical_size   := new_size;
            pe_obj_tab[code32 ].virtual_size    :=
                        pe_obj_tab_all[i].virtual_size
                     +  (pe_obj_tab_all[i].rva-pe_kopf.text_start);
            if not code_vorhanden then
              pe_obj_tab[code32 ].physical_offset := pe_obj_tab_all[i].physical_offset;
            code_vorhanden:=true
          end;
        {VPT: data section found}
        if ((pe_obj_tab_all[i].obj_flags and scnData)=scnData)
           and ((pe_obj_tab_all[i].obj_flags and (scnDiscardable or scnShared))=0)
           and (pe_obj_tab_all[i].rva>=pe_kopf.data_start)
           and (pe_obj_tab_all[i].physical_size>0)
           (* and (pe_obj_tab_all[i].virtual_size>0) *)
           and (pe_obj_tab_all[i].name<>'.idata'#0#0)
           {VPT: many linkers have writeable '.idata'}
           then
          begin
            if pe_kopf.data_start = 0 then {VPT: alink/blink bug}
              begin
                pe_kopf.data_start      := pe_obj_tab_all[i].rva;
                pe_obj_tab[const32].rva := pe_obj_tab_all[i].rva
              end;
            new_size := pe_obj_tab_all[i].physical_size
                     +  (pe_obj_tab_all[i].rva-pe_kopf.data_start);
            ReAllocMem(data_puffer,new_size);
            SeekBST(datei,pe_obj_tab_all[i].physical_offset);
            BlockReadBST
              (
                datei,
                data_puffer^[pe_obj_tab_all[i].rva-pe_kopf.data_start],
                pe_obj_tab_all[i].physical_size
              );
            fill_size := pe_obj_tab_all[i].rva
              - (pe_obj_tab[const32].rva + pe_obj_tab[const32].physical_size);
            if fill_size > 0 then
              FillChar(data_puffer^[pe_obj_tab[const32].physical_size], fill_size, 0);
            pe_obj_tab[const32].physical_size    := new_size;
            pe_obj_tab[const32].virtual_size    :=
                        pe_obj_tab_all[i].virtual_size
                     +  (pe_obj_tab_all[i].rva-pe_kopf.data_start);
            if not data_vorhanden then
              pe_obj_tab[const32 ].physical_offset := pe_obj_tab_all[i].physical_offset;
            data_vorhanden:=true
          end;
        {VPT: import section found}
        if (pe_obj_tab_all[i].rva<=pe_kopf.import_table) and
           (pe_kopf.import_table<pe_obj_tab_all[i].rva+pe_obj_tab_all[i].physical_size) then
          begin
            pe_obj_tab[idata  ].name            := '.idata'#0#0;
            pe_obj_tab[idata  ].virtual_size    := pe_kopf.import_data_size;
            pe_obj_tab[idata  ].rva             := pe_kopf.import_table;
            pe_obj_tab[idata  ].physical_size   := pe_kopf.import_data_size;
            pe_obj_tab[idata  ].physical_offset := pe_obj_tab_all[i].physical_offset
                               + (pe_kopf.import_table - pe_obj_tab_all[i].rva);
            pe_obj_tab[idata  ].obj_flags       := pe_obj_tab_all[i].obj_flags;
            lies(import_puffer,pe_obj_tab[idata]);
            importe_vorhanden:=true
          end;
        {VPT: fixup section found}
        if (pe_obj_tab_all[i].rva<=pe_kopf.fixup) and
           (pe_kopf.fixup<pe_obj_tab_all[i].rva+pe_obj_tab_all[i].physical_size) then
          begin
            pe_obj_tab[reloc  ].name            := '.reloc'#0#0;
            pe_obj_tab[reloc  ].virtual_size    := pe_kopf.fixup_size;
            pe_obj_tab[reloc  ].rva             := pe_kopf.fixup;
            pe_obj_tab[reloc  ].physical_size   := pe_kopf.fixup_size;
            pe_obj_tab[reloc  ].physical_offset := pe_obj_tab_all[i].physical_offset
                               + (pe_kopf.fixup - pe_obj_tab_all[i].rva);
            pe_obj_tab[reloc  ].obj_flags       := pe_obj_tab_all[i].obj_flags;
            lies(relo_puffer,pe_obj_tab[reloc]);
            relo_vorhanden:=true
          end;
        {VPT: resource section found}
        if (pe_obj_tab_all[i].rva<=pe_kopf.resource_table) and
           (pe_kopf.resource_table<pe_obj_tab_all[i].rva+pe_obj_tab_all[i].physical_size) then
          begin
            pe_obj_tab[res    ].name            := '.rsrc'#0#0#0;
            pe_obj_tab[res    ].virtual_size    := pe_kopf.resource_data_size;
            pe_obj_tab[res    ].rva             := pe_kopf.resource_table;
            pe_obj_tab[res    ].physical_size   := pe_kopf.resource_data_size;
            pe_obj_tab[res    ].physical_offset := pe_obj_tab_all[i].physical_offset
                               + (pe_kopf.resource_table - pe_obj_tab_all[i].rva);
            pe_obj_tab[res    ].obj_flags       := pe_obj_tab_all[i].obj_flags;
            lies(reso_puffer,pe_obj_tab[res]);
            resourcen_vorhanden:=true
          end;
        {VPT: bss section found}
        if ((pe_obj_tab_all[i].obj_flags and $80) <> 0) or
           ((pe_obj_tab_all[i].obj_flags and $60) =  0) then
          begin
            new_size := pe_obj_tab_all[i].rva + pe_obj_tab_all[i].virtual_size;
            if pe_obj_tab_all[i].rva < pe_kopf.data_start then
              begin {VPT: add BSS to CODE}
                if new_size > pe_obj_tab[code32].rva + pe_obj_tab[code32].virtual_size then
                  pe_obj_tab[code32].virtual_size := new_size - pe_obj_tab[code32].rva;
              end
            else
              begin {VPT: add BSS to CONST}
                if new_size > pe_obj_tab[const32].rva + pe_obj_tab[const32].virtual_size then
                  pe_obj_tab[const32].virtual_size := new_size - pe_obj_tab[const32].rva;
              end
          end;
        {VPT: To do following sections: EXPORT, DEBUG, TLS}
      end;

    if not (code_vorhanden and data_vorhanden and relo_vorhanden)
     then
      Abbruch('Error: not created with VP 2.x');

    //if pe_obj_tab[idata  ].physical_size>$14 then
    //  WriteLn(textz_warnung_import^);

    if not importe_vorhanden then
      FillChar(pe_obj_tab[idata],SizeOf(pe_obj_typ),0);
    if not resourcen_vorhanden then
      FillChar(pe_obj_tab[res  ],SizeOf(pe_obj_typ),0);

    if module_description='' then
      for i:=0 to pe_kopf.num_obj - 1 do
        begin
          {VPT: search for description, first method}
          if ((pe_obj_tab_all[i].rva<=pe_kopf.image_description) and
             (pe_kopf.image_description<pe_obj_tab_all[i].rva+pe_obj_tab_all[i].physical_size))
             then
            begin
              if pe_kopf.description_size>255 then
                pe_kopf.description_size:=255;
              SeekBST(datei, pe_obj_tab_all[i].physical_offset
                + (pe_kopf.image_description - pe_obj_tab_all[i].rva));
              BlockReadBST(datei,module_description[1],pe_kopf.description_size);
              SetLength(module_description, pe_kopf.description_size);
            end;
          {VPT: search for description, second method}
          if pe_obj_tab_all[i].name = '.desc'#0#0#0 then
            begin
              if pe_obj_tab_all[i].physical_size>255 then
                pe_obj_tab_all[i].physical_size:=255;
              SeekBST(datei, pe_obj_tab_all[i].physical_offset);
              BlockReadBST(datei,module_description[1],pe_obj_tab_all[i].physical_size);
              SetLength(module_description, pe_obj_tab_all[i].physical_size);
            end;
        end;
    {VPT: trim and update description}
    while (module_description<>'') and (module_description[Length(module_description)] = #0) do
      Delete(module_description,Length(module_description),1);
    if module_description='' then
       load_module_description(InFile);

    CloseBST(datei);

  end;

//


procedure rechne_elf;
var
  FilePosition,AddressPosition:longint;
  sekt_z:tSections;
  z1:longint;

  function GetFilePosition: Longint;
  begin
    {AH 16.07.2008 FreeBSD checks for program header alignment of 4, let's align
        everything}
    RoundUp(FilePosition, Alignment);
    Result := FilePosition;
  end;

  function name_speichern(const zk:string):longint;
    begin
      Result:=string_puffer_laenge;
      ReallocMem(string_puffer,string_puffer_laenge+Length(zk)+1);
      Move(zk[1],string_puffer^[string_puffer_laenge],Length(zk));
      Inc(string_puffer_laenge,Length(zk));
      string_puffer^[string_puffer_laenge]:=0; // #0
      Inc(string_puffer_laenge);
    end;

  procedure suche_naechste_passende_dateipostion;
    begin
      if (FilePosition and (PAGESIZE-1))<>(AddressPosition and (PAGESIZE-1)) then
        begin
          FilePosition:=(FilePosition and not (PAGESIZE-1))+PAGESIZE+(AddressPosition and (PAGESIZE-1));
        end;
    end;

  procedure suche_naechste_passende_adressposition;
    begin
      if (FilePosition and (PAGESIZE-1))<>(AddressPosition and (PAGESIZE-1)) then
        begin
          AddressPosition:=(AddressPosition and not (PAGESIZE-1))+PAGESIZE+(FilePosition and (PAGESIZE-1));
        end;
    end;

  procedure naechste_seite;
    begin
      // !!!!! vielleicht lst es ein paar Probleme ...
//      inc(AddressPosition,PAGESIZE);
    end;

  var
    _plt_dateiposition,           _plt_adressposition,
    _rel_plt_dateiposition,       _rel_plt_adressposition,
    _got_dateiposition,           _got_adressposition,
    _dynamic_dateiposition,       _dynamic_adressposition,
    _dynsym_dateiposition,        _dynsym_adressposition,
    _dynsymstr_dateiposition,     _dynsymstr_adressposition,
    _dynsymstrhash_dateiposition, _dynsymstrhash_adressposition,
    resource_dateiposition,       resource_adressposition:longint;

    _plt_laenge,
    _rel_plt_laenge,
    _dynamic_laenge,
    _dynsym_laenge,
    resource_laenge,
    hash_tabelle_laenge:longint;

    luecke_dateiposition,luecke_adressposition,luecke_laenge,_got_laenge:longint;

  procedure Speichere_in_der_DYNAMIC_Tabelle(_d_tag,_d_val:word);
    begin
      Inc(anzahl_dynamic_eintraege);
      ReallocMem(dynamic_speicher,anzahl_dynamic_eintraege*SizeOf(Elf32_Dyn_type));
      with dynamic_speicher^[anzahl_dynamic_eintraege-1] do
        begin
          d_tag:=_d_tag;
          d_val:=_d_val;
        end;
    end;

  (***************************************************************************)

  var
    symboltabelle_gefuellt:longint;

  procedure Symboltabelleneintrag(const _name,_value:word;const _info:byte;const _shindex:smallword);
    begin
      Inc(symboltabelle_gefuellt);
      ReallocMem(symboltabelle,symboltabelle_gefuellt*SizeOf(Elf32_Sym_type));
      with symboltabelle^[symboltabelle_gefuellt-1] do
        begin
          st_name           :=_name;      // Stringtabellenindex
          st_value          :=_value;     // Adressen,..
          st_size           :=0;          // unwichtig
          st_info           :=_info;      // bind/type
          st_other          :=0;          // undefiniert
          st_shndx          :=_shindex;   // auf welche Sektion bezogen
        end;
    end;

  (***************************************************************************)

  var
    symbolstringtabelle_laenge:longint;
    symbolstringtabelle_eintraege:longint;
    _dynamic_st_index,_global_offset_table_st_index:longint;

  function Speichere_in_der_SymbolStringTabelle(const sym_name:string):longint;
    var
      p:pointer;
    begin
      // hier kann noch um ein paar Byte gefeilt werden
      // wenn Teilzeichenketten genutzt werden ..
      Result:=symbolstringtabelle_laenge;
      ReallocMem(symbolstringtabelle,symbolstringtabelle_laenge+Length(sym_name)+1);
      Move(sym_name[1],symbolstringtabelle^[symbolstringtabelle_laenge],Length(sym_name));
      Inc(symbolstringtabelle_laenge,Length(sym_name));
      symbolstringtabelle^[symbolstringtabelle_laenge]:=#0;
      Inc(symbolstringtabelle_laenge);
      Inc(symbolstringtabelle_eintraege);
    end;

  (***************************************************************************)
  (***************************************************************************)

  procedure import_umrechnung;
    type
      pe_Import_Directory_Entry=
        packed record
          IMPORT_FLAGS            :longint;
          TIME_DATE_STAMP         :longint;
          MAJOR_VERSION,
          MINOR_VERSION           :smallword;
          NAME_RVA                :longint;
          IMPORT_LOOKUP_TABLE_RVA :longint;
          //??IMPORT_ADDRESS_TABLE_RVA:longint;
        end;

      temp_import_typ=
        record
          wo,                     // Aufrufer
          sym_index:longint;      // Index in der Symboltabelle
        end;

      temp_import_tabelle_typ=array[1..100000000] of temp_import_typ;

    var
      p,p2,off:longint;
      temp_import_tabelle_z:^temp_import_tabelle_typ;
      temp_import_tabelle_gefuellt:longint;

      z,z2:longint;
      gefunden:boolean;

      plt_index_tabelle:PLongArray;

      weitergerueckt:boolean;

    (************************************************************************)

    procedure Merke_Import(const _wo,_sym_index:longint);
      begin
        Inc(temp_import_tabelle_gefuellt);
        ReallocMem(temp_import_tabelle_z,temp_import_tabelle_gefuellt*SizeOf(temp_import_typ));
        with temp_import_tabelle_z^[temp_import_tabelle_gefuellt] do
          begin
            wo       :=_wo       ;
            sym_index:=_sym_index;
          end;
      end;

    (************************************************************************)

    procedure Erzeuge_Hash_Tabelle;

      type char_z=^char;

      function elf_hash(const _name:char_z):longint;
      // Seite 84
        var
          z:char_z;
          g,h:longint;
        begin
          z:=_name;
          h:=0;
          while z^<>#0 do
            begin
              h:=(h shl 4)+Ord(z^);
              Inc(z);

              if g=(h and $F0000000) then
                h:=h xor (g shr 24);

              h:=h and (not g);
            end;
        end;

      var
        nbucket,nchain:longint;
        z,wert:longint;
      begin
        nbucket:=symboltabelle_gefuellt; // kleiner ?
        nchain :=symboltabelle_gefuellt;

        // die Lnge wurde schon vorher berechnet
        //hash_tabelle_laenge:=4*(1+1+nbucket+nchain);
        GetMem  (hash_tabelle ,hash_tabelle_laenge);
        FillCHar(hash_tabelle^,hash_tabelle_laenge,0);

        hash_tabelle^[0]:=nbucket;
        hash_tabelle^[1]:=nchain;

        // Eintrag 1 ist reserviert ('')
        for z:=2 to nchain do
          begin
            wert:=elf_hash(@symbolstringtabelle^[symboltabelle^[z-1].st_name]);
            wert:=wert mod nbucket;
            if hash_tabelle^[2+wert]=0 then
              // leerer "Eimer" gefunden
              hash_tabelle^[2+(wert mod nbucket)]:=z-1
            else
              begin
                // voller Eimer -> Verkettung
                while hash_tabelle^[2+nbucket+wert]<>0 do
                  wert:=hash_tabelle^[2+nbucket+wert];
                hash_tabelle^[2+nbucket+wert]:=z-1;
              end;
          end;
      end;

    var
      s: tSections;
    begin
      // mit "leerer" Stringtabelle fr Symbole anfangen
      symbolstringtabelle:=nil;
      symbolstringtabelle_eintraege:=0;
      symbolstringtabelle_laenge:=0;
      Speichere_in_der_SymbolStringTabelle(#0);

      _dynamic_st_index            :=Speichere_in_der_SymbolStringTabelle('_DYNAMIC');
      _global_offset_table_st_index:=Speichere_in_der_SymbolStringTabelle('_GLOBAL_OFFSET_TABLE_');

      temp_import_tabelle_gefuellt:=0;
      temp_import_tabelle_z:=nil;

      p:=0;
      while p<pe_obj_tab[idata].physical_size do
        with pe_Import_Directory_Entry(import_puffer^[p]) do
          begin
            if NAME_RVA=0 then
              break;

            DWriteLn('* '+StrPas(@import_puffer^[NAME_RVA-pe_obj_tab[idata].rva]));
            // "shared object"-name als DT_NEEDED Eintrag abspeichern
            Speichere_in_der_DYNAMIC_Tabelle(
              DT_NEEDED,
              Speichere_in_der_SymbolStringTabelle(
                StrPas(@import_puffer^[NAME_RVA-pe_obj_tab[idata].rva])));

            p2:=IMPORT_LOOKUP_TABLE_RVA-pe_obj_tab[idata].rva;
            repeat
              off:=Meml[Ofs(import_puffer^[p2])];
              if off=0 then break;


              if Odd(off shr 31) then
                begin (* import by ordinal *)
                  abbruch(#7'Error: Module imports by ordinal not supported ('+StrPas(@import_puffer^[NAME_RVA-pe_obj_tab[idata].rva])+'.'+Int2Str(off and $7fffffff)+')!');
                end
              else
                begin (* import by name *)
                  Dec(off,pe_obj_tab[idata].rva);
                  DWriteLn('+ $'+Int2Hex(p2+pe_obj_tab[idata].rva+pe_kopf.image_base,8)+' : '+
                          StrPas(@import_puffer^[off+2]));
                  // Funktion und zu Speicherstelle merken wo sie benutzt wird
                  Merke_Import(
                    p2+pe_obj_tab[idata].rva+pe_kopf.image_base,
                    Speichere_in_der_SymbolStringTabelle(
                      StrPas(@import_puffer^[off+2])));
                end;
              Inc(p2,4);
            until false;
            Inc(p,SizeOf(pe_Import_Directory_Entry));
          end;

      //******************************************************************
      // jetzt ist die Anzahl der Importe bekannt

      // Lnge der Stringtabelle auf 32 Bit-Vielfaches bringen
      RoundUp(symbolstringtabelle_laenge,4);

      // zwischen der Interpreter-Sektion und der .text-Sektion ist noch
      // fast 4096 Byte Platz (Adresse und Datei)
      s:=sekt_interpreter;
      luecke_adressposition:=SectionTable[s].sh.sh_addr + SectionTable[s].sh.sh_size;
      luecke_dateiposition :=SectionTable[s].sh.sh_offset+SectionTable[s].sh.sh_size;
      RoundUp(luecke_adressposition,4);
      RoundUp(luecke_dateiposition ,4);
      s:=sekt_text;
      luecke_laenge:=SectionTable[s].sh.sh_addr-luecke_adressposition;

      // VP erzeugt schon soetwas hnliches wie .plt (in .code32)
      // diese Eintrge jmp [.idata+xyz] // $ff $25 (.idata+xyz)
      //                jmp [.idata+xzy]
      //                jmp [.idata+yxz]
      //                ...
      // muessen spter auf die GOT (Global Offset Table) verweisen
      // jetzt wird erst einmal der Bereich und die Anzahl bestimmt
      p2:=pe_obj_tab[code32].physical_size;
      while code_puffer^[p2-2]=0 do dec(p2); (* $0040xxxx *)
      z:=0;
      while (code_puffer^[p2-6]=$ff) and (code_puffer^[p2-6+1]=$25) do
        begin
          Dec(p2,6);
          Inc(z);
        end;

      if z<>temp_import_tabelle_gefuellt then
        abbruch(#7'Error: '+Int2Str(z)+' stubs for imported functions found but '+Int2Str(temp_import_tabelle_gefuellt)+' functions imported');

      // die .dynamic-Tabelle kommt nach dem Datensegment
      _dynamic_adressposition:=AddressPosition;
      _dynamic_dateiposition :=GetFilePosition ;
      // die .dynamic-Tabelle enthlt jetzt schon anzahl_dynamic_eintraege Eintrge
      // auerdem kommen noch hinzu:
      // DT_HASH,DT_STRTAB,DT_SYMTAB,DT_STRSZ,DT_REL,DT_RELSZ,DT_RELENT,DT_NULL
      _dynamic_laenge:=SizeOf(Elf32_Dyn_type)*(anzahl_dynamic_eintraege+8);
      Inc(AddressPosition,_dynamic_laenge);
      Inc(FilePosition ,_dynamic_laenge);

      // die .got kommt nach .dynamic
//  naechste_seite;
      _got_adressposition:=AddressPosition;
      _got_dateiposition :=GetFilePosition ;
      _got_laenge:=(3+temp_import_tabelle_gefuellt)*4;
      Inc(AddressPosition,_got_laenge);
      Inc(FilePosition ,_got_laenge);

      //************************************************
      // .dynsym,.dynstr,.hash,. sind nicht beschreibbar
      RoundUp(AddressPosition,PAGESIZE);
      suche_naechste_passende_adressposition;

      // Symbole in der .dynsym-Tabelle:
      //  1 ""
      // +1 "_DYNAMIC"
      // +1 "_GLOBAL_OFFSET_TABLE_"
      // +temp_import_tabelle_gefuellt
      _dynsym_laenge:=(3+temp_import_tabelle_gefuellt)*SizeOf(Elf32_Sym_type);

      weitergerueckt:=false;
      // Adresse von .dynsym festlegen
      // wenn mglich in der Lcke von .interp bis .text
      if _dynsym_laenge<=luecke_laenge then
        begin
          _dynsym_adressposition:=luecke_adressposition;
          _dynsym_dateiposition :=luecke_dateiposition ;
          Dec(luecke_laenge        ,_dynsym_laenge);
          Inc(luecke_adressposition,_dynsym_laenge);
          Inc(luecke_dateiposition ,_dynsym_laenge);
        end
      else
        begin
          _dynsym_adressposition:=AddressPosition;
          _dynsym_dateiposition :=GetFilePosition ;
          Inc(AddressPosition,_dynsym_laenge);
          Inc(FilePosition ,_dynsym_laenge);
          weitergerueckt:=true;
        end;

      // Adresse von .dynstr festlegen
      // wenn mglich in der Lcke von .interp bis .text
      if symbolstringtabelle_laenge<=luecke_laenge then
        begin
          _dynsymstr_adressposition:=luecke_adressposition;
          _dynsymstr_dateiposition :=luecke_dateiposition ;
          Dec(luecke_laenge        ,symbolstringtabelle_laenge);
          Inc(luecke_adressposition,symbolstringtabelle_laenge);
          Inc(luecke_dateiposition ,symbolstringtabelle_laenge);
        end
      else
        begin
          _dynsymstr_adressposition:=AddressPosition;
          _dynsymstr_dateiposition :=GetFilePosition ;
          Inc(AddressPosition,symbolstringtabelle_laenge);
          Inc(FilePosition ,symbolstringtabelle_laenge);
          weitergerueckt:=true;
        end;

      // Adresse von .hash festlegen
      // Lnge der Hash-Tabelle (siehe erzeuge_hash_tabelle)
      hash_tabelle_laenge:=4*(1+1+2*(3+temp_import_tabelle_gefuellt));
      // wenn mglich in der Lcke von .interp bis .text
      if hash_tabelle_laenge<=luecke_laenge then
        begin
          _dynsymstrhash_adressposition:=luecke_adressposition;
          _dynsymstrhash_dateiposition :=luecke_dateiposition ;
          Dec(luecke_laenge        ,hash_tabelle_laenge);
          Inc(luecke_adressposition,hash_tabelle_laenge);
          Inc(luecke_dateiposition ,hash_tabelle_laenge);
        end
      else
        begin
          _dynsymstrhash_adressposition:=AddressPosition;
          _dynsymstrhash_dateiposition :=GetFilePosition ;
          Inc(AddressPosition,hash_tabelle_laenge);
          Inc(FilePosition ,hash_tabelle_laenge);
          weitergerueckt:=true;
        end;

      _rel_plt_laenge:=temp_import_tabelle_gefuellt*SizeOf(Elf32_Rel_type);
      GetMem  (rel_plt_speicher ,_rel_plt_laenge);
      FillChar(rel_plt_speicher^,_rel_plt_laenge,0);
      // Adresse von .rel.plt festlegen
      // wenn mglich in der Lcke von .interp bis .text
      if _rel_plt_laenge<=luecke_laenge then
        begin
          _rel_plt_adressposition:=luecke_adressposition;
          _rel_plt_dateiposition :=luecke_dateiposition ;
          Dec(luecke_laenge        ,_rel_plt_laenge);
          Inc(luecke_adressposition,_rel_plt_laenge);
          Inc(luecke_dateiposition ,_rel_plt_laenge);
        end
      else
        begin
          _rel_plt_adressposition:=AddressPosition;
          _rel_plt_dateiposition :=GetFilePosition ;
          Inc(AddressPosition,_rel_plt_laenge);
          Inc(FilePosition ,_rel_plt_laenge);
          weitergerueckt:=true;
        end;

      // Reseourcen passen am besten hierher:
      // auch "nur Lesen" (kein Ausfhren,Schreiben)
      if SectionTable[sekt_resource].Exists then
        begin
          resource_laenge:=pe_obj_tab[res].physical_size;
          if resource_laenge<=luecke_laenge then
            begin
              resource_adressposition:=luecke_adressposition;
              resource_dateiposition :=luecke_dateiposition ;
              Dec(luecke_laenge        ,resource_laenge);
              Inc(luecke_adressposition,resource_laenge);
              Inc(luecke_dateiposition ,resource_laenge);
            end
          else
            begin
              resource_adressposition:=AddressPosition;
              resource_dateiposition :=GetFilePosition ;
              Inc(AddressPosition,resource_laenge);
              Inc(FilePosition ,resource_laenge);
              weitergerueckt:=true;
            end;
        end;

      // .plt ist ausfhrbar -> mu auf eine neue Speicherseite
      if weitergerueckt then
        begin
          RoundUp(AddressPosition,PAGESIZE);
          suche_naechste_passende_adressposition;
        end;

      // Lnge der .plt-Sektion ausrechnen
      // .plt hat .plt0 und plt1..plt(Anzahl Importe) Einsprungstellen
      GetMem(plt_index_tabelle,4*(temp_import_tabelle_gefuellt+1));
      // .plt ist kleiner als $10*(anzahl_importe+1)
      GetMem  (plt_speicher ,(temp_import_tabelle_gefuellt+1)*$10);
      FillChar(plt_speicher^,(temp_import_tabelle_gefuellt+1)*$10,0);

      _plt_laenge:=0;
      // .plt0:
      //          push offset got[4]      // $ff $35 (got+4)
      //          jmp got[8]              // $ff $25 (got+8)
      plt_index_tabelle^[0]:=_plt_laenge;
      plt_speicher^[$0]:=$ff;
      plt_speicher^[$1]:=$35;
      pLongint(@plt_speicher^[$2])^:=_got_adressposition+4;
      plt_speicher^[$6]:=$ff;
      plt_speicher^[$7]:=$25;
      pLongint(@plt_speicher^[$8])^:=_got_adressposition+8;
      Inc(_plt_laenge,6+6);

      for z:=1 to temp_import_tabelle_gefuellt do
        begin
          plt_index_tabelle^[z]:=_plt_laenge;
          //.pltx_: (x>=1)
          //      push (x-1)*2                       (A)
          //      jmp .plt0                          (B)

          // je 2 Varianten fr (A) und (B):

          //      $6a shortint((x-1)*8)              (A1)
          //      $eb shortint(.plt0-.pltx_-2-2)     (B1)

          //      $68 longint((x-1)*8)               (A2)
          //      $e9 longint(.plt0-.pltx_-5-5       (B2)

          if (z-1)*8<=127 then
            begin (* A1 *)
              plt_speicher^[_plt_laenge+0]:=$6a;
              plt_speicher^[_plt_laenge+1]:=(z-1)*8;
              Inc(_plt_laenge,2);
            end
          else
            begin (* A2 *)
              plt_speicher^[_plt_laenge+0]:=$68;
              pLongint(@plt_speicher^[_plt_laenge+1])^:=(z-1)*8;
              Inc(_plt_laenge,5);
            end;

          if plt_index_tabelle^[0]-_plt_laenge-2>=-128 then
            begin (* B1 *)
              plt_speicher^[_plt_laenge+0]:=$eb;
              shortint(plt_speicher^[_plt_laenge+1]):=plt_index_tabelle^[0]-_plt_laenge-2;
              Inc(_plt_laenge,2);
            end
          else
            begin (* B2 *)
              plt_speicher^[_plt_laenge+0]:=$e9;
              pLongint(@plt_speicher^[_plt_laenge+1])^:=plt_index_tabelle^[0]-_plt_laenge-5;
              Inc(_plt_laenge,5);
            end;

          // entsprechende "Relokation" erzeugen
          with rel_plt_speicher^[z-1] do
            begin
              r_offset:=_got_adressposition+(3+z-1)*4; // die ersten 3 Eintrge in der GOT sind reserviert
              r_info  :=(z-1+3) shl 8                    // die ersten 3 Eintrge in der Symboltabelle sind reserviert
                       +R_386_JMP_SLOT;
            end;

          pLongint(@plt_speicher^[z*$10+$c])^:=-$10*(z+1);
        end;
      RoundUp(_plt_laenge,4); // auf 32 bit ausrichten

      // jetzt ist die Lnge von .plt bekannt
      // Suche nach freiem Adressraum

      // zwischen der .text-Sektion und der nchsten Seitengrenze ist
      // vielleicht noch etwas Platz (an der Adresse und in der Datei)
      s:=sekt_text;
      luecke_adressposition:=SectionTable[s].sh.sh_addr  +SectionTable[s].sh.sh_size;
      luecke_dateiposition :=SectionTable[s].sh.sh_offset+SectionTable[s].sh.sh_size;
      RoundUp(luecke_adressposition,4);
      RoundUp(luecke_dateiposition ,4);
      luecke_laenge:=luecke_adressposition and (PAGESIZE-1);
      if luecke_laenge=PAGESIZE then
        luecke_laenge:=0;

      // Adresse von .plt festlegen
      // wenn mglich in der Lcke
      if _plt_laenge<=luecke_laenge then
        begin
          _plt_adressposition:=luecke_adressposition;
          _plt_dateiposition :=luecke_dateiposition ;
          Dec(luecke_laenge        ,_plt_laenge);
          Inc(luecke_adressposition,_plt_laenge);
          Inc(luecke_dateiposition ,_plt_laenge);
        end
      else
        begin
          _plt_adressposition:=AddressPosition;
          _plt_dateiposition :=GetFilePosition ;
          Inc(AddressPosition,_plt_laenge);
          Inc(FilePosition ,_plt_laenge);
          weitergerueckt:=true;
        end;

      // jetzt sind wichtige Adressen bekannt und die _DYNAMIC-Tabelle
      // kann vervollstndigt werden

      // Adresse der Hash-Tabelle dem Programmlader verraten
      Speichere_in_der_DYNAMIC_Tabelle(DT_HASH,_dynsymstrhash_adressposition);

      // ........... Symboltabelle-Zeichenketten ...
      Speichere_in_der_DYNAMIC_Tabelle(DT_STRTAB,_dynsymstr_adressposition);

      // ........... Symboltabelle ...
      Speichere_in_der_DYNAMIC_Tabelle(DT_SYMTAB,_dynsym_adressposition);

      // Lnge des Bereiches DT_STRTAB^
      Speichere_in_der_DYNAMIC_Tabelle(DT_STRSZ,symbolstringtabelle_laenge);

      // Adresse des "Relokationen"
      Speichere_in_der_DYNAMIC_Tabelle(DT_REL,_rel_plt_adressposition);
      // Gesamtlnge
      Speichere_in_der_DYNAMIC_Tabelle(DT_RELSZ,_rel_plt_laenge);
      // Lnge eines Eintrages
      Speichere_in_der_DYNAMIC_Tabelle(DT_RELENT,SizeOf(Elf32_Rel_type));

      // Ende der _DYNAMIC-Tabelle
      Speichere_in_der_DYNAMIC_Tabelle(DT_NULL,0);

      //**********************************************************************
      // neue Symboltabelle
      symboltabelle:=nil;
      symboltabelle_gefuellt:=0;

      // Symboltabelleneintraege 0..2
      // 0: "undefiniert"   name,value,info,shindex
      Symboltabelleneintrag(
        0,
        0,
        0,SHN_UNDEF);

      // 1: "_DYNAMIC" - Adresse von ".dynamic"
      Symboltabelleneintrag(
        _dynamic_st_index,
        _dynamic_adressposition,
        (STB_GLOBAL shl 4)+STT_OBJECT,SHN_ABS);

      // 2: "_GLOBAL_OFFSET_TABLE_"
      Symboltabelleneintrag(
        _global_offset_table_st_index,
        _got_adressposition,
        (STB_GLOBAL shl 4)+STT_OBJECT,SHN_ABS);

      // die ersten 3 Eintrge in der Symboltabelle sind schon benutzt
      // die ersten 3 Eintrge in der GOT sind auch reserviert
      // im .plt-Bereich sind die ersten $10 Byte schon benutzt
      GetMem(got_speicher   ,3*4+4*temp_import_tabelle_gefuellt);
      FillChar(got_speicher^,3*4+4*temp_import_tabelle_gefuellt,0);
      got_speicher^[0]:=_dynamic_adressposition;
      got_speicher^[1]:=0;
      got_speicher^[2]:=0;

      // off=Adresse des Code32-Segmentes wo VP schon soetwas wie .plt anlegt
      off:=SectionTable[sekt_text].sh.sh_addr;
      for z:=1 to temp_import_tabelle_gefuellt do
        begin

          // .plt+$10*x+6 ist die Adresse die erstmal zum Programmlader springt
          got_speicher^[z-1+3]:=_plt_adressposition+plt_index_tabelle^[z];

          // suche den Aufruf im code32-Bereich
          gefunden:=false;
          for z2:=1 to temp_import_tabelle_gefuellt do
            if pLongint(@code_puffer^[p2+(z2-1)*6+2])^=temp_import_tabelle_z^[z].wo then
              begin
                gefunden:=true;

                // Code so umstricken, das jetzt auf die GOT verwiesen wird
                pLongint(@code_puffer^[p2+(z2-1)*6+2])^:=_got_adressposition+(z-1+3)*4;

                Symboltabelleneintrag(
                  temp_import_tabelle_z^[z].sym_index,
                  off+p2+(z2-1)*6,
                  (STB_GLOBAL shl 4)+STT_FUNC,SHN_UNDEF);

                break; (* z2 *)
              end;

          if not gefunden then
            Abbruch(#7'Error: caller of imported function '
                   +StrPas(@symbolstringtabelle^[temp_import_tabelle_z^[z].sym_index])
                   +' not found');
        end;
      Erzeuge_Hash_Tabelle;
      // die temp_import_tabelle wird nicht mehr bentigt
      FreeMem(temp_import_tabelle_z);
      anzahl_importe:=temp_import_tabelle_gefuellt;

      FreeMem(plt_index_tabelle);
    end;

  var
    exe_description:string;

  procedure search_description;
    var
      exe_info,locl,z     :longint;
      pas                 :string;
      p,n,e               :string;
      def                 :text;
      zeile               :string;
    begin
      exe_description:='';

      exe_info:=Meml[Ofs(code_puffer^[pe_kopf.entrypoint-pe_obj_tab[code32].rva])+1];
      Dec(exe_info,pe_obj_tab[const32].rva+pe_kopf.image_base);

      locl:=MemL[Ofs(data_puffer^[exe_info])+$c];
      if locl=-1 then
        Exit;

      Inc(locl,12+exe_info);
      if (data_puffer^[locl]<>$3f) then Exit;
      Move(data_puffer^[locl+1],pas,data_puffer^[locl+1]+1);

      for z:=1 to Length(pas) do
        pas[z]:=Char(Ord(pas[z]) xor $aa);

      FSplit(pas,p,n,e);
      Assign(def,p+n+'.def');
      (*$I-*)
      Reset(def);
      (*$I+*)
      if IOResult<>0 then Exit;

      while not Eof(def) do
        begin
          ReadLn(def,zeile);

          while (zeile<>'') and (zeile[1            ] in [' ',#9]) do
            Delete(zeile,1,1);
          while (zeile<>'') and (zeile[Length(zeile)] in [' ',#9]) do
            Dec(zeile[0]);

          for z:=1 to Length('DESCRIPTION') do
            zeile[z]:=UpCase(zeile[z]);

          if Pos('DESCRIPTION',zeile)=1 then
            begin
              Delete(zeile,1,Length('DESCRIPTION'));

              while (zeile<>'') and (zeile[1] in [' ',#9]) do
                Delete(zeile,1,1);

              // '/"
              if (zeile<>'') and (zeile[1            ] in ['''','"']) then
                Delete(zeile,1,1);
              if (zeile<>'') and (zeile[Length(zeile)] in ['''','"']) then
                Dec(zeile[0]);

              exe_description:=zeile;
            end;

        end;
      Close(def);
    end;

  procedure SortSections;
  var
    i, j, k: tSections;
    LastEntry: tSections;
  begin
    {Create sort list: Insert all sections with section headers}
    SectionSortArrayElementCount := low(tSections);
    for i:=low(i) to high(i) do with SectionTable[i] do
      if Exists and Has_sh then
      begin
        SectionSortArray[SectionSortArrayElementCount] := Id;
        Inc(SectionSortArrayElementCount);
      end;

    {Sort the list after p_vaddr and .shstrtab}
    LastEntry := sekt_strings;
    for i:=low(i) to Pred(Pred(SectionSortArrayElementCount)) do
    begin
      for j:=Succ(i) to pred(SectionSortArrayElementCount) do with SectionTable[ SectionSortArray[i] ] do
      begin
        if (ph.p_vaddr > SectionTable[ SectionSortArray[j] ].ph.p_vaddr) and
           (SectionTable[ SectionSortArray[j] ].Id <> LastEntry) or
           (Id = LastEntry) then
        begin
          {i <- j    i -> j}
          k                   := SectionSortArray[i];
          SectionSortArray[i] := SectionSortArray[j];
          SectionSortArray[j] := k;
        end;
      end;
    end;
  end;

  procedure FinishSections;
  var
    i: tSections;
  begin
    {For the next operation all Sections must have the correct index}
    for i:=low(i) to Pred(SectionSortArrayElementCount) do
    with SectionTable[ SectionSortArray[i] ] do
      {Save the sorted list index into the SectionTable[]}
      Index := Ord(i);

    {Fill in missing index values}
    for i:=low(i) to Pred(SectionSortArrayElementCount) do
    with SectionTable[ SectionSortArray[i] ] do
    begin
      {Replace possible index values in the sh_link/sh_info fields}
      sh.sh_link := SectionTable[ tSections(sh.sh_link) ].Index;
      sh.sh_info := SectionTable[ tSections(sh.sh_info) ].Index;
      {Finally save the section header into the save buffer}
      sektionen_kopf[ index ] := sh;
    end;

    {Save the string section header index}
    elf_kopf.e_shstrndx := SectionTable[sekt_strings].Index;
  end;

  var
    oldcodesize          : Longint;
    olddatasize          : Longint;
    s                    : tSections;
    programm_kopf_anzahl : longint;

  procedure kopiere_pk;
    begin
      with SectionTable[s], SectionTable[s].ph, SectionTable[s].sh do
      begin
        {Set the program header data}
        pk2[programm_kopf_anzahl]:=ph;
        Inc(programm_kopf_anzahl);
      end;
    end;

  procedure erweitere_vorhandenen_oder_kopiere_pk;
    var
      t,
      luecke                    :longint;
    begin
      t:=0;
      repeat
        luecke := SectionTable[s].ph.p_vaddr -
                  (pk2[t].p_vaddr + pk2[t].p_filesz);
        if  (SectionTable[s].ph.p_flags = pk2[t].p_flags) and
            (0 <= luecke) and (luecke < 4) and
            (pk2[t].p_type <> PT_INTERP)
         then
          begin
            if SectionTable[s].ph.p_filesz > 0 then
              Inc(pk2[t].p_filesz, luecke + SectionTable[s].ph.p_filesz);
            if SectionTable[s].ph.p_memsz > 0 then
              Inc(pk2[t].p_memsz, luecke + SectionTable[s].ph.p_memsz);
            Exit;
          end;
        Inc(t);
      until t >= programm_kopf_anzahl;
      kopiere_pk;
    end;

  begin {Rechne}
    search_description;

    string_puffer_laenge:=0;
    string_puffer:=nil;

    anzahl_dynamic_eintraege:=0;
    dynamic_speicher:=nil;

    programm_kopf_anzahl:=0;

    //*********************************************************************
    // Lnge eines Symboltabelleneintrages
    Speichere_in_der_DYNAMIC_Tabelle(DT_SYMENT,SizeOf(Elf32_Sym_type));

    //*********************************************************************
    // Code-Bereich aufblhen ? - bringt nichts
    if memorymode>=3+1 then
      begin
        oldcodesize:=pe_obj_tab[code32].physical_size;
        RoundUp(pe_obj_tab[code32].virtual_size ,PAGESIZE);
        RoundUp(pe_obj_tab[code32].physical_size,PAGESIZE);
        ReallocMem(code_puffer,pe_obj_tab[code32].physical_size);

        FillChar(code_puffer^[oldcodesize],pe_obj_tab[code32].physical_size-oldcodesize,0);
      end;

    //*********************************************************************
    // das von VP im Datenbereich angelegte Stacksegment kann gelscht werden
    z1:=Meml[Ofs(code_puffer^[pe_kopf.entrypoint-pe_obj_tab[code32].rva])+5+1];
    Dec(pe_obj_tab[const32].virtual_size,z1);

    //*********************************************************************
    // Nullen am Ende des Datenbereiches kommen in den virtuellen Teil
    with pe_obj_tab[const32] do
      begin
        olddatasize:=physical_size;

        while data_puffer^[physical_size-1-1]=0 do
          Dec(physical_size);
        // auf 32-Bit-Lnge bringen
        RoundUp(physical_size,4);

        if memorymode>=3 then // noch schlimmer aufblhen
          begin
            if virtual_size<physical_size then
              virtual_size:=physical_size;
            RoundUp(virtual_size, PAGESIZE);
            physical_size:=virtual_size;
          end
        else if memorymode>=2 then
          begin
            RoundUp(physical_size,PAGESIZE);
            RoundUp(virtual_size,PAGESIZE);
            Inc(virtual_size,PAGESIZE*10);
          end;
      end;

    ReAllocMem(data_puffer,pe_obj_tab[const32].physical_size);
    if pe_obj_tab[const32].physical_size>olddatasize then
      FillChar(data_puffer^[olddatasize],pe_obj_tab[const32].physical_size-olddatasize,0);

    //*********************************************************************
    // welche tSections werden bentigt ?
    if pe_obj_tab[const32].virtual_size>pe_obj_tab[const32].physical_size then
      SectionTable[sekt_data_b].Exists:=true; (* Variables *)

    if importe_vorhanden then
      begin
        SectionTable[sekt_interpreter   ].Exists:=true;
        SectionTable[sekt_plt           ].Exists:=true;
        SectionTable[sekt_rel_plt       ].Exists:=true;
        SectionTable[sekt_got           ].Exists:=true;
        SectionTable[sekt_dynamic       ].Exists:=true;
        SectionTable[sekt_dynsym        ].Exists:=true;
        SectionTable[sekt_dynsymstr     ].Exists:=true;
        SectionTable[sekt_dynsymstrhash ].Exists:=true;
      end;

    //*********************************************************************
    // Resourceumwandlung
    if resourcen_vorhanden then
      (* if you stop here with an compiler error, then place Use_PE2Elf
         into the Options/Compiler/conditional defines field and recompile all *)
      if not search_resource_anchor(data_puffer,pe_obj_tab[const32].physical_size) then
        begin
          WriteLn('Dropping ',pe_obj_tab[res].physical_size,' bytes unused Resources.');
          resourcen_vorhanden:=false;
          Dispose(reso_puffer);
          reso_puffer:=nil;
          FillChar(pe_obj_tab[res],SizeOf(pe_obj_tab[res]),0);
        end;

    if resourcen_vorhanden then
      begin
        // sekt_resource
        SectionTable[sekt_resource].Exists:=true;
        (* Annahme: viel weniger Platz wird bentigt *)
        GetMem(reso_temp_puffer,(pe_obj_tab[res].physical_size*3) div 2);
        reso_temp_laenge:=0;

        konvertiere_resourcen(reso_puffer,pe_obj_tab[res].physical_size,pe_obj_tab[res].rva,
                              nil,0,
                              reso_temp_puffer,reso_temp_laenge,resource_codepage);

        (* neu in richtiger Gre anfordern *)
        ReAllocMem(reso_puffer,reso_temp_laenge);
        (* Daten Kopieren *)
        Move(reso_temp_puffer^,reso_puffer^,reso_temp_laenge);
        (* temporren Speicher freigeben *)
        Dispose(reso_temp_puffer);
        (* neue Lnge merken *)

        pe_obj_tab[res].virtual_size :=reso_temp_laenge;
        pe_obj_tab[res].physical_size:=reso_temp_laenge;
      end;

    if include_section_header then
    begin
      SectionTable[sekt_strings].Exists:=true;
      SectionTable[sekt_sect_headers].Exists := True;
    end;

    SectionTable[sekt_desc].Exists := module_description <> '';

    //*********************************************************************
    // Count program and section headers
    for s:=low(tSections) to high(tSections) do
    with SectionTable[s], SectionTable[s].ph, SectionTable[s].sh do
      if Exists then
        begin
          {Init flags}
          Has_sh := (sh_type <> SHT_NULL) or (s = sekt_null);
          Has_ph := p_type  <> PT_NULL;
          {Pre-count: the e_phnum is bigger than the final number of ph-entries }
          Inc(elf_kopf.e_shnum, Ord(Has_sh));
          Inc(elf_kopf.e_phnum, Ord(Has_ph));
        end;

    if not include_section_header then
      elf_kopf.e_shnum:=0;

    FilePosition    := 0;
    AddressPosition := 0;

    //*********************************************************************
    // Complete the SectionTable[].ph|sh information
    //***********************************************************************
    for s:=low(s) to high(s) do
    with SectionTable[s], SectionTable[s].ph, SectionTable[s].sh do
      if Exists then
      begin
        if Has_sh then sh_name := name_speichern(Name);
        case s of
          //*********************************************************************
          // ELF header
          sekt_elf_header:
            begin
              sh_addr         :=AddressPosition;
              sh_offset       :=GetFilePosition;
              sh_size         :=SizeOf(elf_kopf);

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;

              Inc(AddressPosition, sh_size);
              Inc(FilePosition   , sh_size);

              AddressPosition := BASEADDRESS-PAGESIZE+GetFilePosition;
            end;
          //*********************************************************************
          // Moduledescription: string from <targetname>.(EDS|DEF|PAS) file
          sekt_desc:
            begin
              module_description := module_description + linebreak + #0;
              sh_addr         :=AddressPosition;
              sh_offset       :=GetFilePosition;
              sh_size         :=Length(module_description);

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;

              Inc(AddressPosition, sh_size);
              Inc(FilePosition   , sh_size);

              RoundUp(AddressPosition,4);
            end;
          //*********************************************************************
          // Program headers
          sekt_prog_headers:
            begin
              sh_addr         :=AddressPosition;
              sh_offset       :=GetFilePosition;
              sh_size         :=elf_kopf.e_phentsize*elf_kopf.e_phnum;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;

              elf_kopf.e_phoff:=GetFilePosition;

              Inc(AddressPosition, sh_size);
              Inc(FilePosition   , sh_size);
//              kopiere_pk;
            end;
          //*********************************************************************
          // Section headers
          sekt_sect_headers:
            begin
              sh_addr         :=0;
              sh_offset       :=GetFilePosition;
              sh_size         :=elf_kopf.e_shentsize*elf_kopf.e_shnum;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;

              elf_kopf.e_shoff:=GetFilePosition;

              Inc(AddressPosition, sh_size);
              Inc(FilePosition   , sh_size);
            end;
          //*********************************************************************
          // NULL section
          sekt_null:
            begin
            end;
          //*********************************************************************
          // Note section
          sekt_note:
            begin
              sh_addr         :=AddressPosition;
              sh_offset       :=GetFilePosition;
              sh_size         :=0;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;

              Inc(AddressPosition, sh_size);
              Inc(FilePosition   , sh_size);

              RoundUp(AddressPosition,4);
//              kopiere_pk;
            end;
          //***********************************************************************
          // Programmlader
          sekt_interpreter:
            begin
              sh_addr         :=AddressPosition;
              sh_offset       :=GetFilePosition;
              sh_size         :=StrLen(programmlader)+1;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;

              Inc(AddressPosition,sh_size);
              Inc(FilePosition   ,sh_size);

              RoundUp(AddressPosition,4);
              Address:=@programmlader;
//              kopiere_pk;
            end;
          //***********************************************************************
          // CODE bearbeiten
          sekt_text:
            begin
              AddressPosition:=pe_obj_tab[code32].rva+pe_kopf.image_base+delta;
              suche_naechste_passende_dateipostion;

              sh_addr         :=AddressPosition;
              sh_offset       :=GetFilePosition;
              sh_size         :=pe_obj_tab[code32].physical_size;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;

              // Eintrittspunkt
              elf_kopf.e_entry:=pe_kopf.entrypoint+pe_kopf.image_base+delta;

              Inc(AddressPosition, sh_size);
              Inc(FilePosition   , sh_size);

              RoundUp(AddressPosition,4);
              Address:=code_puffer;
//              kopiere_pk;
            end;
          //***********************************************************************
          // BSS
          sekt_bss:
            begin
              AddressPosition:=pe_obj_tab[_bss].rva+pe_kopf.image_base+delta;

              sh_addr         :=AddressPosition;
              sh_offset       :=0;
              sh_size         :=pe_obj_tab[_bss].virtual_size;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=0;
              p_memsz         :=sh_size;

              Inc(AddressPosition, sh_size);
              RoundUp(AddressPosition,4);
              Address:=bss_puffer; (* !!!! *)
//              kopiere_pk;
            end;
          //***********************************************************************
          // Konstanten und Daten - Teil 1   .rodata? SHF_ALLOC
          sekt_data_a:
            begin
              AddressPosition:=pe_obj_tab[const32].rva+pe_kopf.image_base+delta;
              suche_naechste_passende_dateipostion;

              sh_addr         :=AddressPosition;
              sh_offset       :=GetFilePosition;
              sh_size         :=pe_obj_tab[const32].physical_size;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;

              Inc(AddressPosition, sh_size);
              Inc(FilePosition   , sh_size);
              Address:=data_puffer;
//              kopiere_pk;
            end;
          //***********************************************************************
          // Konstanten und Daten - Teil 2
          sekt_data_b:
            begin
              sh_addr         :=AddressPosition;
              sh_offset       :=GetFilePosition;
              sh_size         :=pe_obj_tab[const32].virtual_size-pe_obj_tab[const32].physical_size;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=0;
              p_memsz         :=sh_size;

              Inc(AddressPosition, sh_size);
              RoundUp(AddressPosition,4);
              suche_naechste_passende_adressposition;
//              erweitere_vorhandenen_oder_kopiere_pk;
            end;
          //***********************************************************************
          // ".dynamic"
          sekt_dynamic:
            begin
              resource_adressposition:=0;
              if importe_vorhanden then
                import_umrechnung
              else
                begin
                  resource_laenge:=pe_obj_tab[res].physical_size;
                  resource_adressposition:=AddressPosition;
                  resource_dateiposition :=GetFilePosition ;
                  Inc(AddressPosition,resource_laenge);
                  Inc(FilePosition ,resource_laenge);
                end;

              sh_addr         :=_dynamic_adressposition;
              sh_offset       :=_dynamic_dateiposition ;
              sh_size         :=_dynamic_laenge;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;
              Address:=dynamic_speicher;
              p_type:=PT_LOAD;
              erweitere_vorhandenen_oder_kopiere_pk;
              p_type:=PT_DYNAMIC;
//              kopiere_pk;
            end;
          //***********************************************************************
          // ".got"
          sekt_got:
            begin
              sh_addr         :=_got_adressposition;
              sh_offset       :=_got_dateiposition ;
              sh_size         :=_got_laenge;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;
              Address:=got_speicher;
//              erweitere_vorhandenen_oder_kopiere_pk;
            end;
          //***********************************************************************
          // ".dynsym"
          sekt_dynsym:
            begin
              sh_addr         :=_dynsym_adressposition;
              sh_offset       :=_dynsym_dateiposition ;
              sh_size         :=_dynsym_laenge;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;
              Address:=symboltabelle;
//              erweitere_vorhandenen_oder_kopiere_pk;
            end;
          //***********************************************************************
          // ".dynstr"
          sekt_dynsymstr:
            begin
              sh_addr         :=_dynsymstr_adressposition;
              sh_offset       :=_dynsymstr_dateiposition;
              sh_size         :=symbolstringtabelle_laenge;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;
              Address:=symbolstringtabelle;
//              erweitere_vorhandenen_oder_kopiere_pk;
            end;
          //***********************************************************************
          // ".hash"
          sekt_dynsymstrhash:
            begin
              sh_addr         :=_dynsymstrhash_adressposition;
              sh_offset       :=_dynsymstrhash_dateiposition;
              sh_size         :=hash_tabelle_laenge;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;
              Address:=hash_tabelle;
//              erweitere_vorhandenen_oder_kopiere_pk;
            end;
          //***********************************************************************
          // Resourcen
          sekt_resource:
            begin
              sh_addr         :=resource_adressposition;
              sh_offset       :=resource_dateiposition;
              sh_size         :=resource_laenge;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;
              Address:=reso_puffer;
//              kopiere_pk
            end;
          //***********************************************************************
          // ".rel.plt"
          sekt_rel_plt:
            begin
              sh_addr         :=_rel_plt_adressposition;
              sh_offset       :=_rel_plt_dateiposition;
              sh_size         :=_rel_plt_laenge;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;
              Address:=rel_plt_speicher;
//              erweitere_vorhandenen_oder_kopiere_pk;
            end;
          //***********************************************************************
          // ".plt"
          sekt_plt:
            begin
              sh_addr         :=_plt_adressposition;
              sh_offset       :=_plt_dateiposition;
              sh_size         :=_plt_laenge;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=sh_size;
              p_memsz         :=sh_size;
              Address:=plt_speicher;
//              erweitere_vorhandenen_oder_kopiere_pk;
            end;
          //***********************************************************************
          // Sektionsnamen
          sekt_strings:
            begin
              sh_addr         :=0;
              sh_offset       :=GetFilePosition;
              sh_size         :=string_puffer_laenge;

              p_offset        :=sh_offset;
              p_vaddr         :=sh_addr;
              p_paddr         :=sh_addr;
              p_filesz        :=string_puffer_laenge;
              p_memsz         :=string_puffer_laenge;

              Inc(FilePosition ,string_puffer_laenge);
              Inc(AddressPosition,string_puffer_laenge);
              Address:=string_puffer;
            end;
        end;
        {Copy the program header data}
        if Has_ph then erweitere_vorhandenen_oder_kopiere_pk;
        {Save the size of the data to be saved}
        Size := p_filesz;
      end;

    //**********************************************************************
    // Fix the program header number
    elf_kopf.e_phnum := programm_kopf_anzahl;

    //**********************************************************************
    // Output the programm headers
    DWriteln('');
    DWriteln('Program header entries:');
    DWriteLn('# '      +'  '+
             'filepos '+'  '+
             'len     '+'  '+
             'addr    '+'  '+
             'vaddr   '+'  '+
             'len     '+'  '+
             'align   '+'  '+
             'name of section ');
    z1 := 0;
    for s:=low(tSections) to high(tSections) do
    with SectionTable[s], SectionTable[s].ph, SectionTable[s].sh do
      if Exists and Has_ph then
      begin
        DWriteln(Int2Hex(z1,2)+'  '+
                 Int2Hex(sh_offset,8)+'  '+
                 Int2Hex(sh_size,8)+'  '+
                 Int2Hex(sh_addr,8)+'  '+
                 Int2Hex(p_vaddr,8)+'  '+
                 Int2Hex(p_vaddr,8)+'  '+
                 Int2Hex(p_memsz,8)+'  '+
                 Int2Hex(p_align,8)+'  '+
                 name
                );
        Inc (z1);
      end;

    DWriteln('');
    DWriteln('Program header:');
    DWriteLn('# '      +'  '+
             'filepos '+'  '+
             'len     '+'  '+
             'addr    '+'  '+
             'vaddr   '+'  '+
             'len     '+'  '+
             'align   '+'  '+
             'flags   '+'  '+
             'name of section ');
    for z1:=0 to programm_kopf_anzahl-1 do
    with pk2[z1] do
      begin
        DWriteln(Int2Hex(z1,2)+'  '+
                 Int2Hex(p_offset,8)+'  '+
                 Int2Hex(p_filesz,8)+'  '+
                 Int2Hex(p_paddr,8)+'  '+
                 Int2Hex(p_vaddr,8)+'  '+
                 Int2Hex(p_memsz,8)+'  '+
                 Int2Hex(p_align,8)+'  '+
                 Int2Hex(p_flags,8)+'  '+
                 ''
                );
      end;

    //**********************************************************************
    // wenn Resourcen vorhanden sind, mu der Anker in Resource.pas
    // gefunden und angepat werden
    s := sekt_resource;
    with SectionTable[s], SectionTable[s].ph, SectionTable[s].sh do
      if Exists then
      begin
        patch_resource_anchor(data_puffer, sh_addr, sh_size);
      end;

    //**********************************************************************
    // Sort the sections
    SortSections;

    //**********************************************************************
    // Copy the sections into the save buffer and update link/info fields
    FinishSections;
  end;

//

procedure schreibe_elf;
  var
    d2_pos,
    d2_laenge           :longint;

  procedure BlockWrite_d2(const position:longint;var p;const laenge:longint);
    begin
      if d2_pos<>position then
        begin
          d2^.Seek(position);
          if d2^.Status<>stOK then
            abbruch('Stream seek error');
        end;
      d2^.Write(p,laenge);
      if d2^.Status<>stOK then
        abbruch('Stream write error');
      d2_pos:=position+laenge;
      d2_laenge:=Max(d2_laenge,d2_pos);
    end;

  procedure Fuellen(const anzahl:longint);
    var
      null_feld:array[0..PAGESIZE-1] of byte;
    begin
      if anzahl<=0 then Exit;

      DWriteLn('***'+'  '+
               Int2Hex(d2_laenge,8)+'  '+
               Int2Hex(anzahl,8)
               {'********'+'  ',});

      FillChar(null_feld,anzahl,0);
      BlockWrite_d2(d2_laenge,null_feld,anzahl);
    end;

  var
    s                   :tSections;

  begin {schreibe_elf}
    d2:=New(pBufStream,Init(OutFile,stCreate,512*1024));
    if d2^.Status<>stOK then
      abbruch('Error creating '+OutFile+'!');
    d2_pos:=0;
    d2_laenge:=0;

    DWriteln('');
    DWriteln('Section header:');
    DWriteLn('sec'     +'  '+
             'filepos '+'  '+
             'len     '+'  '+
             'addr    '+'  '+
             'vaddr   '+'  '+
             'len     '+'  '+
//             'actpos  '+'  '+
             'attr  '+
             'name of section ');

    for s:=low(tSections) to high(tSections) do
    with SectionTable[s], SectionTable[s].ph, SectionTable[s].sh do
      if Exists and (Address <> nil) then
        begin
          if (Size > 0) and (sh_offset > 0) then
            Fuellen(sh_offset - d2_laenge);

          DWrite  (Int2Hex(Ord(Index),3)+'  '+
                   {Int2Hex(d2_pos,8),'  ',}
                   Int2Hex(sh_offset,8)+'  '+
                   Int2Hex(Size,8)+'  '+
                   Int2Hex(sh_addr,8)+'  '+
                   Int2Hex(p_vaddr,8)+'  '+
                   Int2Hex(p_memsz,8)+'  ');

          if Size > 0 then
            if d2_pos <> p_offset then
              begin
                d2_pos := p_offset;
                d2^.Seek(d2_pos);
                if d2^.Status <> stOK then
                  abbruch('Stream seek error');
              end;

          if (Size > 0) and (p_vaddr > 0) then
            if (d2_pos mod PAGESIZE) <> (p_vaddr mod PAGESIZE) then
              //if memorymode>=2 then//!!!!!
                begin
                  WriteLn('internal bug: filepos=$',
                          Int2Hex(p_offset,8),
                          ' alignment mismatch to v_addr=$',
                          Int2Hex(p_vaddr ,8));
                  RunError(1);
                  end;

//          DWrite(Int2Hex(d2_pos,8) + '  ');

          if (p_flags and PF_R)=PF_R then DWrite('R') else DWrite('-');
          if (p_flags and PF_W)=PF_W then DWrite('W') else DWrite('-');
          if (p_flags and PF_X)=PF_X then DWrite('X') else DWrite('-');
          DWrite('   ');

          DWriteln(Desc);

          BlockWrite_d2(d2_pos, Address^, Size);
        end;

    d2^.Done;

    DWriteln('Success.');
    // memleaks...
  end; {schreibe_elf}

//

function RPos(C: Char; const S: string): LongInt;
begin
  Result := Length(S);
  while (Result > 0) and (S[Result] <> C) do Dec(Result);
end;

procedure show_usage;
  var
    z:word;
  begin
    WriteLn;
    WriteLn('Usage: PE2ELF [options ..] <infile>[.exe] [<outfile>]');
    WriteLn;
    WriteLn(' options:');
    WriteLn('  -v[-] verbose output while converting');
    WriteLn('  -m1 smaller executable');
    WriteLn('  -m2 align to 4096 Byte');
    WriteLn('  -m3 put all uninitialized variables into executable');
//  WriteLn('  -h write section headers');
    WriteLn('  -i<program loader>');
    WriteLn('  -t:<OS/ABI>  valid values are:');
    for z:=Low(OSNames) to High(OSNames) do
      begin
        Write('     -t:',z);
        if z<10 then write(' ');
        Write(' or -t:',OSNames[z],Copy('                ',Length(OSNames[z]),25));
        if Odd(z) then WriteLn;
      end;
    if not Odd(High(OSNames)) then WriteLn;
    WriteLn('  -c:<target codepage> like IBM437, ISO-8859-1; default ISO-8859-1');
    {AH 16.07.2008 new commandline options}
    WriteLn('  -----------');
    WriteLn('  since v1.0:');
    WriteLn('  -----------');
    WriteLn('  -f FreeBSD compatible output (combines several options)');
    WriteLn('  -a:<Num> Align sections & headers in file to <Num> byte boundary (default: 4)');
//    WriteLn('  -o:<Parameter> Use old PE2ELF (< v1.0) operations:');
//    WriteLn('     -o:order Header order: 1. section headers, 2. program headers');
    WriteLn;
    Halt(1);
  end;

{AH 16.07.2008 Parameter code into function for recursive calls}
procedure ParseCommandLineOption( Option: Char; Parameter: String );
var I         : Longint;
    kontrolle : Integer;
begin
  case Option of
    'V': verbose:=(Parameter<>'-');
    'M': begin
           Val(Parameter,memorymode,kontrolle);
           if (memorymode<0) or (memorymode>3) or (kontrolle<>0) then
             show_usage;
         end;
    'I': StrPCopy(programmlader,Parameter);
    'H': include_section_header:=(Parameter<>'-');
    'T': begin
           if Parameter='' then show_usage;

           (* -t:9 ? *)
           Val(Parameter[2],elf_kopf.osabi,kontrolle);
           if kontrolle=0 then Parameter:='';

           (* -t:FreeBSD ? *)
           I:=Low(OSNames);
           while (I<=High(OSNames)) and (Parameter<>'') do
             if Pos(Parameter,OSNames[I])>0 then
               begin
                 elf_kopf.osabi:=I;
                 Parameter:='';
               end
             else
               Inc(I);

           if Parameter<>'' then show_usage;
         end;
    'C': resource_codepage:=Parameter;
    {AH 16.07.2008 new options}
    'F': begin
           ParseCommandLineOption( 'A', '4' );
         end;
    'A': begin
           Val(Parameter,I,kontrolle);
           if (kontrolle = 0) then Alignment := I;
         end;
    'O': begin
           Writeln('NOT IMPLEMENTED!', Parameter);
         end;
  else
    show_usage;
  end;
end;

var
  param         :word;
  para          :string;
  ParaC         :char;

begin
  WriteLn('PE2ELF ' + Version + ' * Veit Kannegieser * 1999.07.28..',datum);
  WriteLn('             * Angels Holocaust * 13.07.2008 - ',ah_date);
  WriteLn('             ' + updated);

  // die Doku sagt '/usr/lib/libc.so.1'#0
  StrCopy(programmlader,'/lib/ld-linux.so.2');

  param:=1;
  repeat
    para:=ParamStr(param);
    if Pos('-',ParamStr(param))=0 then Break;

    Delete(para,1,1);
    if para='' then
      ParaC:=' '
    else
      begin
        ParaC:=UpCase(para[1]);
        Delete(para,1,1);
      end;

    if Pos(':',para)=1 then Delete(para,1,Length(':'));

    ParseCommandLineOption( ParaC, Para );

    Inc(param);
  until false;

  InFile := ParamStr(param);
  if InFile = '' then
    show_usage;

  Inc(param);

  if RPos('.', InFile) = 0 then InFile := InFile + '.exe';
  OutFile := ParamStr(param);
  if OutFile = '' then OutFile := InFile;
  Delete(Outfile, RPos('.', OutFile), 255);

  WriteLn('Converting PE file ''', InFile, ''' to ELF...');

  lade_pe;
  rechne_elf;
  schreibe_elf;
  WriteLn('Done!');
end.
