// used by PE2LE and PE2ELF
unit pe2lety;

interface

(* maybe convert using unit ExeHdr? *)
type
  pe_kopf_typ=
    record
      sig               :array[0..3] of char;
      (* coff Kopf *)
      cpu               :SmallWord;  (* COFF magic, $014C = i386 *)
      num_obj           :SmallWord;  (* number of sections *)
      timdate           :longint;    (* time & date stamp *)
      f_symptr          :longint;    (* COFF: file pointer to symtab *)
      f_nsyms           :longint;    (* COFF: number of symtab entries *)
      nthdrsize         :SmallWord;  (* size of optional header *)
      flags             :SmallWord;
      (* a-out Kopf *)
      magic             :SmallWord;  (* type of file, $010B = executable 32 bit *)
      lmajor            :byte;       (* version stamp *)
      lminor            :byte;
      tsize             :longint;    (* text size in bytes *)
      dsize             :longint;    (* initialized data *)
      bsize             :longint;    (* uninitialized data *)
      entrypoint        :longint;
      text_start        :longint;    (* base of text used for this file *)
      data_start        :longint;    (* base of data used for this file *)
      (* nt Kopf *)
      image_base        :longint;    (* address of specific location in memory *)
      object_align      :longint;    (* section alignment *)
      file_align        :longint;    (* file alignment *)
      osmajor           :SmallWord;       (* minimum OS version required to run *)
      osminor           :SmallWord;
      usermajor         :SmallWord;       (* version of program *)
      userminor         :SmallWord;
      subsysmajor       :SmallWord;       (* minimum subsystem version required to run *)
      subsysminor       :SmallWord;
      win32_version     :longint;    (* immer 0? *)
      image_size        :longint;    (* size of memory to allocate for program *)
      header_size       :longint;    (* size of PE header and section table *)
      file_checksum     :longint;    (* 0 if not used *)
      subsystem         :SmallWord;  (* 1 NATIVE, 2 GUI, 3 CONSOLE, 5 OS/2, 7 POSIX *)
      dll_flags         :SmallWord;  (* immer 0? *)
      stack_reserve_size:longint;
      stack_commit_size :longint;
      heap_reserve_size :longint;
      heap_commit_size  :longint;
      loader_flags      :longint;    (* immer 0? *)
      interestings      :longint;    (* immer 16? *)
      (* rva Tabelle *)
      export_table      :longint;    (* export directory RVA *)
      export_data_size  :longint;
      import_table      :longint;    (* import directory RVA *)
      import_data_size  :longint;
      resource_table    :longint;    (* resource directory RVA *)
      resource_data_size:longint;
      exception_table   :longint;    (* Windows CE ONLY *)
      exception_table_size:longint;
      security_table    :longint;    (* certificates directory PHYSICAL offset *)
      security_size     :longint;
      fixup             :longint;    (* base relocation table RVA *)
      fixup_size        :longint;
      debug             :longint;    (* debug directory RVA *)
      debug_directories :longint;
      image_description :longint;    (* description string RVA *)
      description_size  :longint;
      machine_specific  :longint;    (* MIPS ONLY *)
      machine_specific_size:longint;
      thread_local_storage:longint;  (* TLS directory RVA *)
      tls_size          :longint;
      load_config_table :longint;    (* load configuration directory RVA *)
      load_config_size  :longint;
      bound_import_table:longint;    (* bound import directory PHYSICAL offset *)
      bound_import_size :longint;
      IAT_table         :longint;    (* IAT RVA *)
      IAT_size          :longint;
      delay_import_table:longint;    (* delay load import descriptors RVA *)
      delay_import_size :longint;
      com_runtime_table :longint;    (* COM+ Runtime descriptor RVA *)
      com_runtime_size  :longint;
      unknown_table     :longint;
      unknown_size      :longint;
    end;

  pe_obj_typ=
    record
      name                      :array[0..7] of char;
      virtual_size              :longint;
      rva                       :longint;
      physical_size             :longint;
      physical_offset           :longint;
      s_relptr                  :longint;   (* COFF: file ptr to relocations *)
      s_lnnoptr                 :longint;   (* COFF: file ptr to line numbers *)
      s_nreloc                  :SmallWord; (* COFF: number of relocation entries *)
      s_nlnno                   :SmallWord; (* COFF: number of line number entries *)
      obj_flags                 :longint;
    end;

  pe_obj_typen=(code32,_bss,const32,idata,reloc,res);

  le_kopf_typ=
    record
      sig                       :array[0..1] of char;
      byte_order                :byte;
      word_order                :byte;
      executable_format_level   :longint;
      cpu_typ                   :SmallWord;
      target_os                 :SmallWord;
      module_version            :longint;
      module_type               :longint;
      number_of_memory_pages    :longint;
      cs_obj_nummer             :longint;
      eip_obj_ofs               :longint;
      ss_obj_nummer             :longint;
      esp_obj_ofs               :longint;
      memory_page_size          :longint;
      bytes_on_last_page        :longint; (* LX: page offset shift *)
      fixup_section_size        :longint;
      fixup_section_checksum    :longint;
      loader_section_size       :longint;
      loader_section_checksum   :longint;
      offset_of_object_table    :longint;
      object_table_entries      :longint;
      object_page_map_table_offset:longint;
      object_iterate_data_map_offset:longint;
      resource_table_offset     :longint;
      resource_table_entries    :longint;
      resident_names_table_offset:longint;
      entry_table_offset        :longint;
      module_directives_table_offset:longint;
      Module_Directives_entries :longint;
      Fixup_page_table_offset   :longint;
      Fixup_record_table_offset:longint;
      imported_modules_name_table_offset:longint;
      imported_modules_count    :longint;
      imported_procedures_name_table_offset:longint;
      perpage_checksum_table_offset:longint;
      data_pages_offset         :longint;
      preload_page_count        :longint;
      nonresident_names_table_offset:longint;
      nonresident_names_table_length:longint;
      nonresident_names_checksum:longint;
      automatic_data_object     :longint;
      debug_information_offset  :longint;
      debug_information_length:longint;
      preload_instance_pages_number:longint;
      demand_instance_pages_number:longint;
      extra_heap_allocation     :longint;
      stacksize                 :longint; (* size of stack OS/2 only *)
      reserved                  :array[$b0..$b7] of byte;
      vxd_winresoff             :longint;
      vxd_winreslen             :longint;
      vxd_device_ID             :SmallWord;
      vxd_DDK_version           :SmallWord;
    end;

  le_obj_typ=
    record
      virtual_size_in_bytes     :longint;
      relocation_base_address   :longint;
      object_flags              :longint;
        (* 0      readable
           1      writable
           2      executable
           3      resource
           4      discardable
           5      shared
           6      preloaded
           7      invalid
           8-9    type
                  00 normal
                  01 zero-filled
                  10 resident
                  11 resident and contiguous
           10     resident and long-lockable
           11     reserved
           12     16:16 alias required
           13     "BIG" (Huge: 32-bit)
           14     conforming
           15     "OBJECT_I/O_PRIVILEGE_LEVEL"
           16-31  reserved                       *)

      page_map_index            :longint;
      page_map_entries          :longint;
      reserved                  :array[$14..$17] of byte;
    end;

  rec24bit=array[0..2]of byte;

  le_page_map_entry_typ=
    record
      index_into_fixup_table    :rec24bit;
        (* 0000h if no relocation info *)
      type_                     :byte;
        (* 00h hard copy in file, 03h some relocation needed *)
    end;

  lx_page_map_entry_typ=
    record
      index_into_fixup_table    :longint;
        (* 0000h if no relocation info *)
      data_size                 :SmallWord;
      type_                     :SmallWord;
        (* 00h hard copy in file, 03h some relocation needed *)
    end;

  le_page_map_entry_feld_typ       =array[1..80000] of le_page_map_entry_typ;
  le_page_map_entry_feld_z_typ     =^le_page_map_entry_feld_typ;
  lx_page_map_entry_feld_typ       =array[1..80000] of lx_page_map_entry_typ;
  lx_page_map_entry_feld_z_typ     =^lx_page_map_entry_feld_typ;

  Fixup_page_table_typ          =array[1..80000] of longint;

  Fixup_page_table_z_typ        =^Fixup_page_table_typ;

  pe_relo_typ                   =
    record
      page_rva                  :longint;
      block_size                :longint;
      fixup                     :array[1..8000] of SmallWord;
    end;
  pe_relo_z_typ=^pe_relo_typ;

  le_relo_typ                   =
    record
      type_                     :byte;
       (* bits 7-4: modifier (0001 single, 0011 multiple)
          bits 3-0: type
                  0000 byte offset
                  0010 word segment
                  0011 16-bit far pointer (DWORD)
                  0101 16-bit offset
                  0110 32-bit far pointer (PWORD)
                  0111 32-bit offset
                  1000 near call or jump, WORD/DWORD based on seg attrib *)
      linear_executable_fixup_type:byte;
       (* 7      ordinal is BYTE rather than WORD
          6      16-rather than 8-object number/module ordinal
          5      addition with DWORD rather than WORD
          4      relocation info has size with new two bytes at end
          3      reserved (0)
          2      set if add to destination, clear to replace destination
          1-0    type
                 00 internal fixup
                 01 external fixup, imported by ordinal
                 10 external fixup, imported by name
                 11 internal fixup via entry table                       *)
      number_of_items           :byte;
      object_number             :byte;
      ofs_t                     :array[0..255] of SmallWord;
    end;
  le_relo_z_typ                 =^le_relo_typ;

  exe_kopf_typ                  =
    record
      sig                       :array[0..1] of char;
      mod512,
      page512,
      relo_anz,
      kopfgroesse,
      min,max,
      _ss,
      _sp,
      pruefsumme,
      _ip,
      _cs,
      relo_ofs,
      ov_nummer                 :SmallWord;
      rest                      :array[$1c..$3b] of byte;
      x_exe_zeiger              :longint;
    end;

  info_block_typ=
    record
      ofs,
      phys_laenge,
      log_laenge                :longint;
    end;

  abschnitte=(startcode,
              resource,
              code,
              bss,
              daten,
              alles_zusammen);

  info_block_tabelle_typ        =array[startcode..alles_zusammen] of info_block_typ;

  startcode_typ=(startcode_kein,
                 startcode_watcom,
                 startcode_relopack,
                 startcode_debug,
                 startcode_relopack__code_schon_vorhanden);

const
  scnCode = $60000020; { CODE + EXECUTE + READ }
  scnData = $C0000040; { INITIALIZED_DATA + READ + WRITE }
  scnBss  = $40000080; { UNINITIALIZED_DATA + READ }
  scnDiscardable = $02000000; { DISCARDABLE }
  scnShared=       $10000000; { SHARED }

implementation

end.
