// /p:1 - small entrypoint, just a signature for DOS4GW
procedure start1;assembler;
  {$S-}{&Frame-}{&Uses None}
  asm
    // Length
    dd @start1_end
    dd @start1_begin

  @start1_begin:
    db $e9
    dd 'EIP '
    db 'WATCOM',0
  @start1_end:

  end;

// /p:2 - entrypoint that can apply relocations
procedure start2;assembler;
  {$S-}{&Frame-}{&Uses None}
  asm
    // Length
    dd @start2_VAR5_end
    dd @start2_VAR5_begin

  @start2_VAR5_begin:

    call @pop_edx
  @call_pop_edx:
    db 'WATCOM',0
  @pop_edx:
    pop edx
    sub edx,'5555'      // @call_pop_edx-@start2_virtuapl_pascal_code...

    mov eax,'EIP '
    add eax,edx
    push eax

      mov esi,'RELO'
      add esi,edx
      push esi          // remember starting address of memory area to clean

        cld
        lodsd
        lea edi,[eax+edx]
  @relocation_loop:
        add [edi],edx
        add edi,4

        sub eax,eax
        lodsb

        // more than 253 bytes difference?
        cmp al,254
        jb @small_delta
        // end of relocation data?
        // (al=255)
        ja @relocations_done

        // >253
        lodsd
  @small_delta:
        add edi,eax
        jmp @relocation_loop

  @relocations_done:

      pop edi           // FillChar(Mem[edi],ecx,0)
      push es           // ES (PSP) store
        push ds
        pop es          // ES:=data area

        sub eax,eax
        mov ecx,'VAR0'
        push ecx
          shr ecx,2     // div 4
          rep stosd
        pop ecx
        and ecx,3       // mod 4
        rep stosb
      pop es

      ret               // to program entry point

    db '****'           // supply a save place for a fixup that points to [1:0]
    db '****'           // some more space to avoid fixups on a page break.

    //
    //*****************************************************************

  @start2_VAR5_end:
  end;

// /p:3 - like /p:2 but used for IDE debug support
procedure start3;assembler;
  {$S-}{&Frame-}{&Uses None}
  const
    target_code_ofs     =$00401000;     // Win32 exe load address
    vpddid_data_len     =4+80;          // longint+80 chars
  asm
    // Length
    dd @start3_end
    dd @start3_begin

  @start3_begin:

    call @pop_edx
  @call_pop_edx:
    db 'WATCOM*VPDB',0
    db '\DEV\VPDDID$$',0
  @pop_edx:
    pop ebp
    sub ebp,5                                   // @call_pop_edx-@start3_begin

    push es                                     // PSP selector sichern
      push ds
      pop es

      // install check
      mov eax,$00003d00
      lea edx,[ebp+17]                          // '\DEV\VPDDID$$'
      int $21
      jnc @device_driver_opened_successful

      // previously this would fail hard. changed it to a warning message
      //call @pop_error_halt
      call @pop_message_vpddid_missing
      db 'VPDDID.SYS/EXE nicht installiert.',13,10
      db 'bersetze mich ohne deb_link oder starte mich aus der Entwicklungsumgebung!',13,10
      db 'VPDDID.SYS/EXE not installed.',13,10
      db 'Recompile me without unit deb_link or start me from IDE!',13,10
      db '$'
    @pop_message_vpddid_missing:
      mov ah,9
      pop edx
      int $21
      lea edx,[esp-5000]                        // buffer

      mov [edx].Longint,0                       // do not have XMS
      jmp @driver_done

    @device_driver_opened_successful:
      mov ebx,eax                               // handle

      mov eax,$4400                             // read device information word
    //mov ebx,ebx                               // handle
      int $21

      or dl,$20                                 // +5    raw (binary) mode
      mov dh,0                                  // dh must be zero

      mov eax,$4401                             // set device information word
    //mov ebx,ebx                               // handle
      int $21

      // read pipe name
      mov eax,$00003f00                         // read
    //mov ebx,ebx                               // handle
      mov ecx,vpddid_data_len                   // number of bytes to read
      lea edx,[esp-5000]                        // buffer
      int $21
      jc @read_error
      cmp eax,vpddid_data_len
      je @read_successful
    @read_error:
      call @pop_error_halt
      db 'Lesefehler von VPDDID.SYS/EXE!',13,10
      db 'read error from VPDDID.SYS/EXE!',13,10
      db '$'
    @read_successful:
      mov ah,$3e                                // close device
      int $21
    @driver_done:

      mov ebx,'CODE'                            // Begin of Code
      lea esi,[ebp+ebx]
      mov edi,[edx]                             // longint from VPDDID.SYS
      // memory size. will be zero on Windows DOS session
      or edi,edi
      mov edi,target_code_ofs                   // use XMS
      jnz @have_xms
      mov edi,esi                               // use same memory
    @have_xms:

      mov eax,'EIP '
      add eax,edi
      sub eax,ebx
      push eax

        mov ecx,'RELO'
        push ecx        // remember starting address of memory area to clean
          push edi      // new code start

            // size of code+data=ecx-ebx=end data-begin old code
            sub ecx,ebx

            cld

            // copy program code and data
            push ecx
              shr ecx,2
              rep movsd
            pop ecx
            and ecx,3
            rep movsb

            mov [edx+0],edi                     // remember end of new code+data
            mov [edx+4],esi                     // remember end of old code+data

          pop edi                               // get back begin of new code
        pop esi                                 // get back relocations offset

        mov eax,edi                             // ebx:=Delta=@begin new-@begin old
        sub eax,ebx
        mov ebx,eax

        add esi,ebp                             // esi:=@reloactions

        cld
        lodsd                                   // first relocation address
        lea edi,[eax+ebx]
    @relocation_loop:
        // resource anchor?
        cmp [edi+8].longint,'OSER'              // RESO
        jne @not_resource_pointer

        mov [edi]        ,ebp
        add [edi].Longint,'REOF'                // resource offset
        jmp @continue_resource_pointer

      @not_resource_pointer:                    // else:
        add [edi].Longint,ebx                   // apply relocations

      @continue_resource_pointer:
        add edi,4                               // resources do not overlap, do better coding

        sub eax,eax                             // get next data
        lodsb

        cmp al,254                              // end of relocation data?
        ja @relocations_done
        jne @small_delta                        // less than 253 bytes difference?

        lodsd
    @small_delta:
        add edi,eax                             // edi:=next relocation address
        jmp @relocation_loop

    @relocations_done:

        // zerofill variables area
        sub eax,eax
        mov ecx,'VAR0'                          // size of area
        push ecx

          mov edi,[edx+0]                       // end of new code/data location
          push ecx
            shr ecx,2                           // div 4
            rep stosd
          pop ecx
          and ecx,3                             // mod 4
          rep stosb

        pop ecx

        mov edi,[edx+4]                         // end of old code/data location
        push ecx
          shr ecx,2                             // div 4
          rep stosd
        pop ecx
        and ecx,3                               // mod 4
        rep stosb

      pop eax                                   // VP entrypoint
    pop es                                      // restore PSP selector
    jmp eax                                     // to entrypoint

  (*************************************************************************)

  @pop_error_halt:
    pop edx
    mov ah,$09
    int $21
    mov ah,$08
    int $21
    mov ax,$4cff
    int $21

  (*************************************************************************)

    db '****'           // supply a save place for a fixup that points to [1:0]

  @start3_end:

  end;

